import os
import platform
import asyncio
from curl_cffi import requests
from typing import List

try:
    import psutil
except ImportError as e:
    raise ImportError(
        "Missing psutil dependencies. "
        "Please install: pip install psutil"
    ) from e

class FDUtil:
    """Cross-platform file descriptor / handle utility class"""

    @staticmethod
    def get_max_fd():
        """
        Get the maximum number of files/handles that the current process can open.
        Windows: CRT _getmaxstdio
        Linux/macOS: resource.RLIMIT_NOFILE
        """
        if platform.system() == "Windows":
            import ctypes
            msvcrt = ctypes.cdll.msvcrt
            if hasattr(msvcrt, "_getmaxstdio"):
                return msvcrt._getmaxstdio()
            else:
                # Return default heuristic value
                return 512
        else:
            import resource
            soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
            return soft

    @staticmethod
    def get_used_fd():
        """
        Get the number of file descriptors / handles currently used by the process.
        Windows: psutil.Process().num_handles()
        Linux/macOS: len(/proc/self/fd)
        """
        if platform.system() == "Windows":
            if psutil is None:
                raise RuntimeError("psutil is required on Windows to get used handles")
            p = psutil.Process()
            return p.num_handles()
        else:
            try:
                return len(os.listdir(f"/proc/self/fd"))
            except Exception:
                # Fallback: /proc folder not available, cannot determine
                return -1

    @staticmethod
    def print_fd_info():
        """Print the current process's max FD and used FD count"""
        max_fd = FDUtil.get_max_fd()
        used_fd = FDUtil.get_used_fd()
        print(f"[FDUtil] Max FD: {max_fd}, Used FD: {used_fd}")

async def measure_connection_memory(url="http://127.0.0.1:8002", test_connections=50):
    process = psutil.Process()
    mem_before = process.memory_info().rss

    sessions: List[requests.AsyncSession] = []
    for _ in range(test_connections):
        s = requests.AsyncSession()
        sessions.append(s)
        await s.get(url)

    mem_after = process.memory_info().rss
    per_connection_bytes = (mem_after - mem_before) / test_connections
    available_mem = psutil.virtual_memory().available
    safety_factor = 0.3
    safe_connections = int((available_mem * safety_factor) / per_connection_bytes)

    for s in sessions:
        await s.close()

    return per_connection_bytes, safe_connections

if __name__ == "__main__":
    FDUtil.print_fd_info()

    if platform.system().startswith("win"):
        # asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

    per_conn, safe_max = asyncio.run(measure_connection_memory())
    print(f"[Estimate] Each connection approx: {per_conn/1024:.2f} KB")
    print(f"[Estimate] Safe max concurrent connections: {safe_max}")


import asyncio
from .base import BaseSpider
from ..core.downloader.internet.request import HttpRequest
from ..hooks import spiders_hooks
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..crawler import Crawler

class RedisSpider(BaseSpider):
    name = "redisSpider"
    redis_key = "redis_key"

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            settings=crawler.settings,
            run_py_dir=crawler.run_py_dir,
            stop_event=crawler.stop_event,
            kafkaManager=crawler.kafkaManager,
            session_id="",
            hooks=spiders_hooks(crawler),
        )

    async def start(self, *args, **kwargs):
        while not self.stop_event.is_set():
            get_req_task = asyncio.create_task(self.hooks.scheduler.get_start_req(spider=self))
            stop_task = asyncio.create_task(self.stop_event.wait())
            done, pending = await asyncio.wait(
                {get_req_task, stop_task},
                return_when=asyncio.FIRST_COMPLETED
            )
            if stop_task in done:
                get_req_task.cancel()
                try:
                    await get_req_task
                except asyncio.CancelledError:
                    pass
                break
            if get_req_task in done:
                data = get_req_task.result()
                if not data:
                    await asyncio.sleep(1)
                    continue
                request = await self.make_request_from_data(data)
                if request:
                    yield request

    # By default, only a URL is expected. If data is in JSON format, this method should be overridden in subclasses.
    async def make_request_from_data(self, data: bytes):
        return HttpRequest(
            url=data.decode('utf-8'),
            method="GET",
            headers=self.settings.DEFAULT_HEADERS,
            cookies=self.settings.DEFAULT_COOKIES,
            proxies=self.settings.PROXIES,
            timeout=self.settings.TIMEOUT,
            dont_filter=self.settings.DONT_FILTER,
            callback=self.parse, 
            errback=self.errRet
        )


import asyncio, json
from pathlib import Path
from ..core.downloader.internet.request import HttpRequest
from ..hooks import spiders_hooks
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..core.downloader.internet.response import HttpResponse
    from ..exceptions import Failure
    from ..crawler import Crawler
    from ..hooks.spiders import SpidersHooks
    from ..settings import SettingsInfo
    from ..mq.kafka import KafkaManager

class BaseSpider(object):
    name = "cffiSpider"
    robot_scheme = "https"
    allowed_domains = []

    def __init__(self, settings=None, run_py_dir="", stop_event=None, kafkaManager=None, session_id="", hooks=None, *args, **kwargs):
        self.settings: "SettingsInfo" = settings
        self.run_py_dir: Path = run_py_dir
        self.stop_event: asyncio.Event = stop_event
        self.kafkaManager: "KafkaManager" = kafkaManager
        self.session_id = session_id # If not set, all will share the default session
        self.hooks: "SpidersHooks" = hooks
        
        # Whether to load the JS method; place it under the project's root js_path
        self.ctx_dict = {}
        if self.settings.JS_PATH:
            import execjs, os
            if isinstance(self.settings.JS_PATH, str):
                js_path = Path(self.settings.JS_PATH)
            else:
                js_path = self.run_py_dir / "js_path"
            js_files = os.listdir(js_path)
            for js_file in js_files:
                single_js_file_path = js_path / js_file
                self.ctx_dict["".join(js_file.split(".")[:-1])] = execjs.compile(open(single_js_file_path, encoding='utf-8').read())

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            settings=crawler.settings,
            run_py_dir=crawler.run_py_dir,
            stop_event=crawler.stop_event,
            kafkaManager=crawler.kafkaManager,
            session_id="",
            hooks=spiders_hooks(crawler),
        )

    def use_execjs(self, ctx_key: str="", funcname: str="", params: tuple=()) -> str:
        # funcName = funcname + str(params)
        funcName = f"{funcname}({','.join(json.dumps(p) for p in params)})"
        encrypt_words = self.ctx_dict[ctx_key].eval(funcName)
        return encrypt_words
    
    async def parse(self, response: "HttpResponse"):
        raise NotImplementedError("parse is no defined.")
    
    async def errRet(self, failure: "Failure"):
        print(str(failure))
        yield None

class Spider(BaseSpider):
    start_urls = []
        
    async def start(self, *args, **kwargs):
        for url in self.start_urls:
            yield HttpRequest(
                session_id=self.session_id,
                url=url,
                method="GET",
                headers=self.settings.DEFAULT_HEADERS,
                cookies=self.settings.DEFAULT_COOKIES,
                proxies=self.settings.PROXIES,
                timeout=self.settings.TIMEOUT,
                dont_filter=self.settings.DONT_FILTER,
                callback=self.parse, 
                errback=self.errRet,
            )


try:
    import zstandard as zstd
except ImportError as e:
    raise ImportError(
        "Missing zstandard dependencies. "
        "Please install: pip install zstandard"
    ) from e

from typing import Optional

class ZstdFactory:
    @staticmethod
    def encode(data: bytes, level: int = 3, dict_data: Optional[bytes] = None) -> bytes:
        cctx = zstd.ZstdCompressor(level=level, dict_data=dict_data)
        return cctx.compress(data)

    @staticmethod
    def decode(data: bytes, dict_data: Optional[bytes] = None) -> bytes:
        dctx = zstd.ZstdDecompressor(dict_data=dict_data)
        try:
            return dctx.decompress(data)
        except Exception as e:
            raise ValueError("Unable to zstd-decompress.") from e

__all__ = [
    "ZstdFactory"
]



"""
Portions of this module are derived from the open-source project
'blackboxprotobuf' (https://github.com/nccgroup/blackboxprotobuf),
originally developed by NCC Group Plc and distributed under the MIT License.

The source has been refactored and simplified for internal use within
scrapy_cffi, keeping only the commonly used APIs: `encode_message` and
`decode_message`.

MIT License (c) 2018–2023 NCC Group Plc
"""
from .api import *


# Key Notes:
# Single-threaded asyncio design: All operations are expected to run within the same event loop to avoid concurrency problems.
# Reference counting: Ensures sessions are not closed while still in use.
# Group sessions: Supports grouping multiple session IDs under one logical group for random selection and batch closing.
# WebSocket management: Each session can maintain multiple WebSocket connections, managed individually.
# Safe async cleanup: Uses a background task (_reaper_loop) to close sessions when no longer needed.

import asyncio, hashlib, random
from curl_cffi import requests
from tenacity import AsyncRetrying, stop_after_attempt, wait_fixed, retry_if_exception_type
from typing import Union, Dict, Set, TYPE_CHECKING, Literal, Optional, List
from .downloader.internet import MediaRequest
from ..utils import create_uniqueId, run_with_timeout, safe_call
if TYPE_CHECKING:
    from logging import Logger
    from ..crawler import Crawler
    from ..settings import SettingsInfo
    from curl_cffi.requests import Response
    from .downloader.internet import HttpRequest, WebSocketRequest
    from ..mq.kafka import KafkaManager
    from ..models.api import WebSocketMsg

class CloseSignal:
    def __init__(
        self, 
        session_id: str, 
        websocket_end_for_key: Union[str, Literal[False], None]=False, 
        websocket_end_for_url: Union[str, Literal[False], None]=False, 
        session_end=False
    ):
        self.session_id = session_id
        self.websocket_end_for_key = websocket_end_for_key
        self.websocket_end_for_url = websocket_end_for_url
        self.session_end = session_end

    def __repr__(self):
        return f"<CloseSignal session_id={self.session_id} ws_end={self.websocket_end_for_url} sess_end={self.session_end}>"

class WebSocketEntry:
    """
    Represents a single WebSocket connection identified by its URL (including parameters) and the underlying WebSocket object.
    Manages reference counting for shared usage, supports graceful close on last release or explicit end marking, and handles listener task cancellation and connection cleanup.
    """
    def __init__(
        self, 
        logger, 
        end_tag: str, 
        url: str, 
        task: asyncio.Task, 
        queue: asyncio.Queue,
        ping_data: "WebSocketMsg"=None,
        ping_interval: float=15.0
    ):
        self.logger: "Logger" = logger
        self.end_tag: str = end_tag
        self.url: str = url
        self.task: asyncio.Task = task
        self.websocket: requests.websockets.WebSocket = None
        self.queue: asyncio.Queue = queue
        self.stop_event: asyncio.Event = asyncio.Event()
        self.stop_event.clear()

        self.ref_count = 0
        self.marked_end = False
        self._closed = False
        self._close_lock = asyncio.Lock()

        self._ping_task: asyncio.Task = None
        if ping_data is not None:
            self._ping_task = asyncio.create_task(self._ping_loop(ping_data, ping_interval))

    def acquire(self):
        self.ref_count += 1

    def release(self):
        self.ref_count -= 1
        if self.marked_end and self.ref_count <= 0:
            asyncio.create_task(self.close())

    def mark_end(self):
        self.marked_end = True
        if self.ref_count <= 0:
            asyncio.create_task(self.close())

    async def _ping_loop(self, ping_data: "WebSocketMsg", interval):
        try:
            while not self.stop_event.is_set() and not self.marked_end:
                await asyncio.sleep(interval)
                try:
                    if self.websocket:
                        await safe_call(self.websocket.send, ping_data.data, flags=ping_data.flags)
                    else:
                        await asyncio.sleep(0)
                except Exception as e:
                    self.logger.warning(f"[WebSocketEntry] Ping failed for {self.url}: {e}")
                    break
        except asyncio.CancelledError:
            raise

    async def close(self):
        try:
            async with self._close_lock:
                if self._closed:
                    return
                self._closed = True

                self.stop_event.set()
                await self.queue.put(self.end_tag)

                if self._ping_task:
                    self._ping_task.cancel()
                    try:
                        await self._ping_task
                    except asyncio.CancelledError:
                        pass

                if self.task:
                    self.task.cancel()
                    try:
                        await self.task
                    except asyncio.CancelledError:
                        self.logger.debug(f"[WebSocketEntry] Listener task cancelled for {self.url}")
                    except Exception as e:
                        self.logger.warning(f"[WebSocketEntry] Listener task raised exception for {self.url}: {e}")

                ws_curl = getattr(getattr(self.websocket, "curl", None), "_curl", None)
                if ws_curl is None:
                    self.logger.debug(f"[WebSocketEntry] websocket already released for {self.url}")
                    return

                if hasattr(self.websocket, "close"):
                    try:
                        await safe_call(self.websocket.close)
                    except Exception as e:
                        self.logger.warning(f"[WebSocketEntry] websocket.close() error for {self.url}: {e}")
        except BaseException as e:
            self.logger.error(f"[WebSocketEntry] Error closing websocket for {self.url}: {e}")

class WebSocketPool:
    """
    Manages all WebSocket connections under a single session.
    Keeps a dictionary of WebSocketEntry objects keyed by the MD5 hash of their URL.
    Supports initialization, acquisition, release, marking end, retrieval, and cleanup of WebSocket connections.
    """
    def __init__(self, logger=None):
        self._pool: Dict[str, WebSocketEntry] = {}
        self.logger: "Logger" = logger

    def _key(self, url: str) -> str:
        return hashlib.md5(url.encode("utf-8")).hexdigest()

    def init_websocket(
        self, 
        end_tag: str, 
        url: str, 
        task: asyncio.Task, 
        queue: asyncio.Queue, 
        ping_data: "WebSocketMsg"=None, 
        ping_interval: float=15.0,
    ) -> str: # return websocket_id
        key = self._key(url)
        if key not in self._pool:
            self._pool[key] = WebSocketEntry(
                logger=self.logger, 
                end_tag=end_tag, 
                url=url, 
                task=task, 
                queue=queue,
                ping_data=ping_data,
                ping_interval=ping_interval
            )
        return key
    
    def set_websocket(self, url: str, websocket: requests.websockets.WebSocket) -> str: # return websocket_id
        key = self._key(url)
        if key not in self._pool:
            raise ValueError(f'WebSocketEntry has not been initialized yet.')
        self._pool[key].websocket = websocket
        return key
    
    def get_from_key(self, key: str) -> Optional[WebSocketEntry]:
        return self._pool.get(key)

    def get_from_url(self, url: str) -> Optional[WebSocketEntry]:
        key = self._key(url)
        return self.get_from_key(key)
    
    def acquire_from_key(self, key: str):
        websocket_entry = self.get_from_key(key)
        if websocket_entry:
            websocket_entry.acquire()

    def acquire_from_url(self, url: str):
        websocket_entry = self.get_from_url(url)
        if websocket_entry:
            websocket_entry.acquire()

    def release_from_key(self, key: str):
        websocket_entry = self.get_from_key(key)
        if websocket_entry:
            websocket_entry.release()

    def release_from_url(self, url: str):
        websocket_entry = self.get_from_url(url)
        if websocket_entry:
            websocket_entry.release()

    def mark_end_from_key(self, key: str):
        websocket_entry = self.get_from_key(key)
        if websocket_entry:
            websocket_entry.mark_end()

    def mark_end_from_url(self, url: str):
        websocket_entry = self.get_from_url(url)
        if websocket_entry:
            websocket_entry.mark_end()

    def remove(self, key: str) -> Optional[WebSocketEntry]:
        entry = self._pool.pop(key, None)
        return entry

    async def close_all(self):
        for entry in list(self._pool.values()):
            await entry.close()
        self._pool.clear()

class SessionWrapper:
    """
    Represents a single HTTP/WS session identified by a session ID.
    Wraps an asynchronous HTTP session (curl_cffi.requests.AsyncSession) and maintains a WebSocketPool for that session.
    Supports configuring session-level cookies, retry policy, performing HTTP and WebSocket requests with retries, and closing connections.
    """
    def __init__(self, stop_event: asyncio.Event, settings: "SettingsInfo"=None, cookies: Dict=None, kafkaManager: "KafkaManager" = None):
        self.stop_event = stop_event
        self.settings = settings
        from ..utils import init_logger
        self.logger = init_logger(log_info=self.settings.LOG_INFO, logger_name=__name__)
        if kafkaManager:
            from ..utils import KafkaLoggingHandler
            kafka_handler = KafkaLoggingHandler(kafka=kafkaManager, stop_event=self.stop_event).create_fmt(self.settings)
            self.logger.addHandler(kafka_handler)

        self.session = requests.AsyncSession()
        self.websocket_pool: WebSocketPool = WebSocketPool(logger=self.logger)
        self.default_cookies = cookies or self.settings.DEFAULT_COOKIES
        self.update_session_cookies(self.default_cookies)
        self._lock = asyncio.Lock()

        retry_times = self.settings.MAX_REQ_TIMES
        retry_delay = self.settings.DELAY_REQ_TIME
        self.retryer = AsyncRetrying(
            stop=stop_after_attempt(retry_times),
            wait=wait_fixed(retry_delay),
            retry=retry_if_exception_type((ConnectionError, TimeoutError, OSError)),
            reraise=True
        )

    def _build_request_args(self, request: "HttpRequest") -> Dict:
        args = {
            "url": request.url,
            "headers": request.headers,
            "cookies": request.cookies,
            "proxies": request.proxies,
            "timeout": request.timeout,
            "allow_redirects": request.allow_redirects,
            "max_redirects": request.max_redirects,
            "verify": request.verify,
            "impersonate": request.impersonate,
            "ja3": request.ja3,
            "akamai": request.akamai,
        }
        if request.data:
            args["data"] = request.data
        args.update({k: v for k, v in request.kwargs.items() if k != "json"})
        return args
    
    async def do_request(self, session: requests.AsyncSession, request: Union["HttpRequest", "WebSocketRequest"], is_ws=False):
        async for attempt in self.retryer:
            with attempt:
                if is_ws:
                    return await self.ws_connect_once(session, request)
                else:
                    return await self.do_request_once(session, request)
                
    async def media_req(self, session: requests.AsyncSession, request: MediaRequest):
        all_file_data = []
        part_byte_start = 0
        part_byte_end = request.single_part_size
        single_part_response = None
        while not self.stop_event.is_set():
            if request.media_size < part_byte_end: # The size of the last segment = total file size - the starting index of the next segment to obtain the file bytes
                part_byte_end = request.media_size - part_byte_start
            else:
                part_byte_end = part_byte_start + request.single_part_size
            
            range_key = request.find_header_key("Range")
            range_key = range_key if range_key else "Range"
            request.headers[range_key] = f"bytes={part_byte_start}-{part_byte_end}"
            single_part_response: "Response" = await session.request(
                method=request.method, 
                **self._build_request_args(request)
            )
            single_part_data = single_part_response.content
            all_file_data.append(single_part_data)
            part_byte_start = part_byte_end + 1
            if part_byte_start >= request.media_size:
                media_data = b''.join(all_file_data)
                single_part_response.content = media_data
                break
        return single_part_response
    
    async def do_request_once(self, session: requests.AsyncSession, request: "HttpRequest"):
        if isinstance(request, MediaRequest):
            return await self.media_req(session=session, request=request)
        
        method: Literal["GET", "POST", "PUT", "DELETE", "OPTIONS", "HEAD", "TRACE", "PATCH"] = request.method
        raw_response = await session.request(
            method=method,
            **self._build_request_args(request)
        )
        return raw_response
    
    async def ws_connect_once(self, session: requests.AsyncSession, request: "WebSocketRequest"):
        websocket: "requests.websockets.WebSocket" = await session.ws_connect(url=request.url, 
            headers=request.headers, 
            cookies=session.cookies.get_dict(), 
            proxies=request.proxies, 
            timeout=request.timeout,
            allow_redirects=request.allow_redirects,
            max_redirects=request.max_redirects,
            verify=request.verify,
            impersonate=request.impersonate,
            ja3=request.ja3,
            akamai=request.akamai,
            **request.kwargs
        )
        # Automatic pinging is built-in, but curl_cffi lacks `ping_data` config,
        # so manual protocol-level ping frames cannot be sent.
        return websocket
    
    def get_websocket(self, url: str) -> WebSocketEntry:
        return self.websocket_pool.get_from_url(url)
    
    def init_websocket(self, url: str, task: asyncio.Task, queue: asyncio.Queue, ping_data: bytes=None, ping_interval: float=15.0) -> str: # return websocket_id
        return self.websocket_pool.init_websocket(end_tag=self.settings.WS_END_TAG, url=url, task=task, queue=queue, ping_data=ping_data, ping_interval=ping_interval)

    def set_websocket(self, url: str, websocket: requests.websockets.WebSocket) -> str: # return websocket_id
        return self.websocket_pool.set_websocket(url=url, websocket=websocket)

    def update_session_cookies(self, cookies_dict: Dict):
        for ck, val in cookies_dict.items():
            self.session.cookies.set(ck, val)

    async def session_close(self):
        await self.session.close()

    async def close_websocket(self, key: str):
        entry = self.websocket_pool.get_from_key(key)
        if entry:
            await entry.close()
            self.websocket_pool.remove(key)

class SessionManager:
    """
    The central manager for all sessions running within a single-threaded asyncio event loop.
    Maintains a mapping from session IDs to SessionWrapper instances and groups of session IDs.
    Tracks reference counts for each session to manage usage, marks sessions as ended when tasks complete, and queues sessions for safe asynchronous cleanup via a background reaper loop.
    Provides methods to get/create sessions, batch register sessions with cookies, acquire/release sessions references, mark sessions as ended, and close sessions/groups safely without concurrency issues.
    """
    def __init__(self, stop_event=None, settings=None, kafkaManager=None):
        self._default_session_id = create_uniqueId()
        self._sessions: Dict[str, SessionWrapper] = {self._default_session_id: None}
        self._group_sessions: Dict[str, List[str]] = {}

        self.stop_event: asyncio.Event = stop_event
        self.settings: "SettingsInfo" = settings
        self.kafkaManager: "KafkaManager" = kafkaManager

        # Tracks the current reference count (usage) for each session_id. Format: {session_id: count}
        self._ref_counts: Dict[str, int] = {} 

        # Marks session_ids whose tasks have completed and are eligible for release.
        # Uses a set to ensure idempotency, added via user calls to mark_end.
        self._end_flags: set[str] = set() 
        
        # Queue of session_ids that have met the conditions for closure,
        # to be processed asynchronously by the background _reaper_loop coroutine.
        self._close_queue: asyncio.Queue = asyncio.Queue()

        # A deduplication set to prevent the same session_id from being added multiple times to the close queue.
        self._pending_close_set: Set[str] = set()
        from ..utils import init_logger
        self.logger = init_logger(log_info=self.settings.LOG_INFO, logger_name=__name__)
        if self.kafkaManager:
            from ..utils import KafkaLoggingHandler
            kafka_handler = KafkaLoggingHandler(kafka=self.kafkaManager, stop_event=self.stop_event).create_fmt(self.settings)
            self.logger.addHandler(kafka_handler)

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event, 
            settings=crawler.settings,
            kafkaManager=crawler.kafkaManager,
        )

    def debug_sessions(self):
        self.logger.debug(f"[SessionManager] Current sessions: {list(self._sessions.keys())}")
        self.logger.debug(f"[SessionManager] Reference counts: {self._ref_counts}")
        self.logger.debug(f"[SessionManager] End flags: {self._end_flags}")
        self.logger.debug(f"[SessionManager] Pending close queue: {list(self._pending_close_set)}")
            
    def start(self):
        if not hasattr(self, "_reaper_task") or self._reaper_task.done():
            self._reaper_task = asyncio.create_task(self._reaper_loop())

    def get_or_create_session(self, session_id: str, cookies: Dict=None) -> SessionWrapper:
        if not session_id:
            session_id = self._default_session_id

        if session_id in self._group_sessions:
            actual_session_ids = self._group_sessions[session_id]
            if not actual_session_ids:
                raise ValueError(f"[SessionManager] Group session '{session_id}' has no valid session members.")
            session_id = random.choice(actual_session_ids)

        wrapper = self._sessions.get(session_id)
        if wrapper:
            if cookies:
                wrapper.update_session_cookies(cookies)
            return wrapper

        wrapper = SessionWrapper(stop_event=self.stop_event, settings=self.settings, cookies=cookies, kafkaManager=self.kafkaManager)
        self._sessions[session_id] = wrapper
        return wrapper
    
    def register_sessions_batch(self, user_cookies: Dict[str, Dict], group_id: Optional[str] = None) -> str:
        if not user_cookies:
            return

        group_id = group_id or create_uniqueId()
        session_ids = []

        for session_id, cookies in user_cookies.items():
            if session_id not in self._sessions:
                wrapper = SessionWrapper(stop_event=self.stop_event, settings=self.settings, cookies=cookies, kafkaManager=self.kafkaManager)
                self._sessions[session_id] = wrapper
                session_ids.append(session_id)
            else:
                self.logger.info(f"[SessionManager] Session {session_id} already exists, skipped.")

        self._group_sessions[group_id] = session_ids
        self.logger.debug(f"[SessionManager] Registered group '{group_id}' with sessions: {session_ids}")
        return group_id
    
    async def close_group_sessions(self, group_id: str):
        session_ids = self._group_sessions.pop(group_id, [])
        for session_id in session_ids:
            self.mark_end(session_id)

    def freeze(self):
        self._frozen = True
    
    def is_default_session(self, session_id: str) -> bool:
        return session_id == self._default_session_id

    def acquire(self, session_id: str):
        # self.debug_sessions()
        if self.is_default_session(session_id) or not session_id:
            return
        self._ref_counts[session_id] = self._ref_counts.get(session_id, 0) + 1
        # self.debug_sessions()

    def release(self, session_id: str):
        # self.debug_sessions()
        if self.is_default_session(session_id) or not session_id:
            return
        if session_id not in self._ref_counts:
            self.logger.warning(f"[SessionManager] Release called on unacquired session_id: {session_id}")
            return
        self._ref_counts[session_id] -= 1

        if (session_id in self._end_flags) and (self._ref_counts[session_id] <= 0) and (session_id not in self._pending_close_set):
            self._close_queue.put_nowait(session_id)
            self._pending_close_set.add(session_id)
        # self.debug_sessions()

    def mark_end_single(self, session_id):
        self._end_flags.add(session_id)
        
        ref_count = self._ref_counts.get(session_id, 0)
        if ref_count <= 0 and (session_id not in self._pending_close_set):
            self._close_queue.put_nowait(session_id)
            self._pending_close_set.add(session_id)

    def mark_end(self, session_id: str):
        if self.is_default_session(session_id) or (not session_id):
            return
        
        if session_id in self._group_sessions:
            for it in self._group_sessions[session_id]:
                self.mark_end_single(it)
        else:
            _group_session = self._group_sessions.copy()
            for it in self._group_sessions:
                if session_id in self._group_sessions[it]:
                    _group_session[it].remove(session_id)
                    break
            self._group_sessions = _group_session
            self.mark_end_single(session_id)

    async def _reaper_loop(self):
        try:
            while not self.stop_event.is_set():
                session_id = await run_with_timeout(self._close_queue.get, stop_event=self.stop_event, timeout=0.5)
                try:
                    await self._safe_close(session_id)
                except Exception as e:
                    self.logger.error(f"[SessionManager] Error closing session {session_id}: {e}")
                finally:
                    self._pending_close_set.discard(session_id)
                    self._close_queue.task_done()
            raise asyncio.CancelledError()
        except asyncio.CancelledError:
            raise

    async def _safe_close(self, session_id: str):
        if self.is_default_session(session_id) or (not session_id):
            return
        wrapper = self._sessions.pop(session_id, None)
        if wrapper:
            self.logger.debug(f"[SessionManager] Closing session: {session_id}")
            await wrapper.websocket_pool.close_all()
            await wrapper.session_close()

    def get_session_cookies(self, session_id: str) -> Union[Dict, None]:
        ret_cookies = {}
        if session_id in self._group_sessions:
            session_ids = self._group_sessions[session_id]
            for it in session_ids:
                wrapper = self._sessions.get(it)
                if wrapper:
                    ret_cookies[it] = wrapper.session.cookies.get_dict()
        else:
            wrapper = self._sessions.get(session_id)
            if wrapper:
                ret_cookies = {session_id: wrapper.session.cookies.get_dict()}
        return ret_cookies

    async def close_all(self) -> None:
        await asyncio.gather(*[self._safe_close(session_id) for session_id in list(self._sessions.keys())])
        self._sessions.clear()
        self._ref_counts.clear()
        self._end_flags.clear()
        self._pending_close_set.clear()
        while not self._close_queue.empty():
            self._close_queue.get_nowait()
            self._close_queue.task_done()


import asyncio, time, inspect
from ...utils import async_context_factory, safe_call
from typing import Tuple, TYPE_CHECKING, List, Callable
# from ...utils import run_with_timeout
from .internet import *
from ...exceptions import DownloadError
from ...extensions import signals, SignalInfo
if TYPE_CHECKING:
    from ...crawler import Crawler
    from ...settings import SettingsInfo
    from ...extensions import SignalManager
    from ..sessions import SessionManager, SessionWrapper
    from ...mq.kafka import KafkaManager

class Downloader:
    def __init__(
        self, 
        stop_event: asyncio.Event=None, 
        settings: "SettingsInfo"=None, 
        sessions: "SessionManager"=None, 
        sessions_lock=None, 
        signalManager: "SignalManager"=None,
        kafkaManager: "KafkaManager"=None
    ):
        self.stop_event = stop_event
        self.settings = settings
        from ...utils import init_logger
        self.logger = init_logger(log_info=self.settings.LOG_INFO, logger_name=__name__)
        if kafkaManager:
            from ...utils import KafkaLoggingHandler
            kafka_handler = KafkaLoggingHandler(kafka=kafkaManager, stop_event=self.stop_event).create_fmt(self.settings)
            self.logger.addHandler(kafka_handler)

        self.sessions = sessions
        self.sessions_lock = sessions_lock
        self.signalManager = signalManager
        # Set the maximum concurrency limit for requests
        self.sem_ctx = async_context_factory(
            max_tasks=self.settings.MAX_CONCURRENT_REQ,
            semaphore_cls=asyncio.Semaphore if not self.settings.USE_STRICT_SEMAPHORE else None
        )

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event,
            settings=crawler.settings,
            sessions=crawler.sessions,
            sessions_lock=crawler.sessions_lock,
            signalManager=crawler.signalManager,
            kafkaManager=crawler.kafkaManager,
        )
    
    async def fetch_http(self, request: HttpRequest, callback: Callable) -> asyncio.Task:
        try:
            self.signalManager.send(signal=signals.request_reached_downloader, data=SignalInfo(signal_time=time.time(), request=request))
            wrapper: "SessionWrapper" = self.sessions.get_or_create_session(
                session_id=request.session_id,
                cookies=request.cookies
            )
            raw_response = None
            async with self.sem_ctx():
                raw_response = await wrapper.do_request(session=wrapper.session, request=request)

            if raw_response:
                response = HttpResponse(
                    session_id=request.session_id,
                    raw_response=raw_response,
                    meta=request.meta,
                    dont_filter=request.dont_filter,
                    callback=request.callback,
                    errback=request.errback,
                    desc_text=request.desc_text,
                    request=request
                )
                self.logger.debug(f'request for {request.url} result -> status_code: {response.status_code}')
                self.signalManager.send(signal=signals.response_received, data=SignalInfo(signal_time=time.time(), request=request, response=response))
                await callback(response=response, request=request)
            else:
                self.logger.warning(f'HTTP request timed out or got no response: {request.url}')
        except asyncio.CancelledError as e:
            raise
        except Exception as e:
            result = DownloadError(request=request, exception=e)
            self.logger.error(str(result))
            await callback(response=result, request=request)
        finally:
            async with self.sessions_lock:
                self.sessions.release(session_id=request.session_id)

    async def cancel_ws_tasks(self, tasks: List[asyncio.Task]):
        for t in tasks:
            if not t.done():
                t.cancel()
        for t in tasks:
            try:
                await t
            except Exception as e:
                self.logger.debug(f"Downloader Task cancelled or failed: {e}")

    async def downloaderWebSocketListener(self, websocket_event: asyncio.Event, wrapper: "SessionWrapper", request: WebSocketRequest, queue: asyncio.Queue):
        try:
            async with self.sem_ctx():
                websocket = await wrapper.do_request(session=wrapper.session, request=request, is_ws=True)
                websocket_id = wrapper.set_websocket(url=request.url, websocket=websocket)
                if request.send_message: # Sending requests is supported during the initial WebSocket handshake
                    # await run_with_timeout(websocket.send, request.send_message, stop_event=self.stop_event)
                    for msg in request.send_message:
                        await safe_call(websocket.send, msg.data, flags=msg.flags)

                while (not self.stop_event.is_set()) and (not websocket_event.is_set()):
                    try:
                        if inspect.iscoroutinefunction(websocket.recv):
                            recv_task = asyncio.create_task(websocket.recv())
                        else:
                            recv_task = asyncio.create_task(asyncio.to_thread(websocket.recv))
                        wait_task = asyncio.create_task(websocket_event.wait())
                        stop_task = asyncio.create_task(self.stop_event.wait())
                        tasks = [recv_task, wait_task, stop_task]
                        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
                        if recv_task in done:
                            msg = recv_task.result()
                            # if msg[0] in [b'\x03\xe8', b'\x03\xe8Bye', b'\x03\xf3keepalive ping timeout']: # Predefined termination messages as per protocol convention
                            #     break
                            
                            response = WebSocketResponse(
                                session_id=request.session_id,
                                websocket_id=websocket_id,
                                msg=msg,
                                meta=request.meta,
                                callback=request.callback,
                                errback=request.errback,
                                desc_text=request.desc_text,
                                request=request
                            )
                            self.signalManager.send(signal=signals.response_received, data=SignalInfo(signal_time=time.time(), request=request, response=response))
                            await queue.put(response)
                        else:
                            await self.cancel_ws_tasks(tasks=tasks)
                            break
                    except asyncio.CancelledError:
                        await self.cancel_ws_tasks(tasks=tasks)
                        raise
                    except Exception as e:
                        result = DownloadError(exception=e, request=request)
                        self.logger.error(str(result))
                        await queue.put(result)
                        # if "initializer for ctype" in str(e):
                        #     self.logger.info(f"WebSocket connection {request.url} has already been closed. Exiting listener.")
                        break
        except asyncio.CancelledError:
            raise
        except Exception as e:
            # self.logger.error(f"DownloadError：{e}")
            result = DownloadError(exception=e, request=request)
            self.logger.error(str(result))
            try:
                await queue.put(result)
            except Exception as e:
                self.logger.warning(f"fetch Queue put failed: {e}")
        finally:
            websocket_event.set()
            await queue.put(self.settings.WS_END_TAG)

    async def fetch_websocket(self, wrapper: "SessionWrapper", request: WebSocketRequest) -> Tuple[asyncio.Task, asyncio.Queue, asyncio.Event]:
        self.signalManager.send(signal=signals.request_reached_downloader, data=SignalInfo(signal_time=time.time(), request=request))
        websocket_event = asyncio.Event()
        websocket_event.clear()
        queue = asyncio.Queue()
        task = asyncio.create_task(self.downloaderWebSocketListener(websocket_event=websocket_event, wrapper=wrapper, request=request, queue=queue))
        websocket_id = wrapper.init_websocket(url=request.url, task=task, queue=queue, ping_data=request.ping_data, ping_interval=request.ping_interval)
        return task, queue, websocket_event


import sys
from pathlib import Path
from scrapy_cffi.utils import get_run_py_dir
from scrapy_cffi.settings import SettingsInfo

def create_settings(spider_path, env_path=None, used_redis=False, used_rabbitmq=False, used_kafka=False, *args, **kwargs):
    if env_path:
        env_file = Path(env_path)
        if env_file.exists():
            return env_to_settings(env_file, SettingsInfo)

    settings = SettingsInfo()
    settings.TIMEOUT = 30
    settings.SPIDERS_PATH = spider_path
    settings.EXTENSIONS_PATH = "extensions.CustomExtension"
    settings.ITEM_PIPELINES_PATH = ["pipelines.CustomPipeline2", "pipelines.CustomPipeline1"]
    settings.DOWNLOAD_INTERCEPTORS_PATH = {
        "interceptors.CustomDownloadInterceptor1": 300,
        "interceptors.CustomDownloadInterceptor2": 200,
    }
    settings.JS_PATH = str(get_run_py_dir() / "js_path") # can be a custom path string, or True to use the default: get_run_py_dir() / "js_path"

    if sys.platform.startswith("win"):
        settings.MAX_GLOBAL_CONCURRENT_TASKS = None
        settings.MAX_CONCURRENT_REQ = None

    # settings.DUPEFILTER = "scrapy_cffi.dupefilter.BloomDupeFilter" # In-memory Bloom filter deduplication
    # settings.DUPEFILTER = "scrapy_cffi.dupefilter.api.RedisBloomDupeFilter" # Enable Redis Bloom filter deduplication

    if used_rabbitmq:
        settings.SCHEDULER_PERSIST = True
        settings.SCHEDULER = "scrapy_cffi.scheduler.RabbitMqScheduler"
        settings.REDIS_INFO.URL = "redis://127.0.0.1:6379" # Used for request deduplication
        settings.RABBITMQ_INFO.URL = "amqp://guest:guest@localhost"
    elif used_redis:
        settings.SCHEDULER = "scrapy_cffi.scheduler.RedisScheduler" # Starting the Redis scheduler requires configuring Redis information
        settings.REDIS_INFO.URL = "redis://127.0.0.1:6379"

    if used_kafka:
        settings.KAFKA_INFO.URL = "localhost:9092"

    # settings.LOG_INFO.LOG_FILE = "demo.log"

    # Register a C extension module
    # settings.CPY_EXTENSIONS.DIR = "cpy_extensions"
    # from scrapy_cffi.models import CPYExtension
    # settings.CPY_EXTENSIONS.RESOURCES = [
    #     CPYExtension(module_name="bloom")
    # ] # Usage after injected: import bloom

    # settings.LOG_INFO.LOG_ENABLED = False # Disable logging entirely
    return settings

if __name__ == "__main__":
    from scrapy_cffi.utils.envConfig import settings_to_env, env_to_settings
    from scrapy_cffi.utils import get_run_py_dir
    
    spider_path = str(get_run_py_dir() / "spiders")
    env_path = str(get_run_py_dir() / ".env.dev")

    settings: SettingsInfo = create_settings(spider_path)
    settings_to_env(settings, env_path)

    settings: SettingsInfo = env_to_settings(env_path, SettingsInfo)
    print(settings.model_dump())


from typing import Protocol, TYPE_CHECKING, Dict
if TYPE_CHECKING:
    from ..extensions import SignalInfo

class SignalHooks(Protocol):
    def send(self, signal: object, data: "SignalInfo") -> None: ...

class _SessionHooks(Protocol):
    def mark_end(self, session_id: str) -> None: ...

    def get_session_cookies(self, session_id: str) -> Dict: ...

class SessionHooks(Protocol):
    def get_session_cookies(self, session_id: str) -> Dict: ...

class _PipelinesHooks(Protocol):
    session: _SessionHooks

class PipelinesHooks(Protocol):
    session: SessionHooks
    signals: SignalHooks


import math, secrets, time, hashlib
from typing import List, Union

def do_sha1(data: bytes) -> str:
    return hashlib.sha1(data).hexdigest()

def create_uniqueId():
    origin_array = [int(time.time()), math.floor(secrets.randbits(32) / 4294967296 * 4294967296)]
    value = (origin_array[0] << 32) + origin_array[1]
    if value >= 2**63:
        value -= 2**64
    return str(value)

def do_otp(secret: str, counter: int=None, timestamp_10: int=None) -> str:
    try:
        import pyotp
        secret_clean = secret.replace(" ", "")
        if counter is not None:
            return pyotp.HOTP(secret_clean).at(counter)
        else:
            totp = pyotp.TOTP(secret_clean)
            if timestamp_10:
                return totp.at(timestamp_10)
            return totp.now()
    except ImportError as e:
        raise ImportError(
            "Missing pyotp dependencies. "
            "Please install: pip install pyotp"
        ) from e

def get_node(nodes: List[str], fingerprint: Union[str, bytes]) -> str:
    try:
        import jump
        if isinstance(fingerprint, str):
            fingerprint = fingerprint.encode("utf-8")
        if not isinstance(fingerprint, bytes):
            raise ValueError("fingerprint must Union[str, bytes]")
        key_int = int(hashlib.md5(fingerprint).hexdigest(), 16)
        idx = jump.hash(key_int, len(nodes))
        return nodes[idx]
    except ImportError as e:
        raise ImportError(
            "Missing jump dependencies. "
            "Please install: pip install jump-consistent-hash"
        ) from e


# utils/wire.py
import binascii
from ..schema import VarInt
from ..exceptions import EncoderException, DecoderException
from typing import Any, Tuple

def encode_uvarint(value: Any) -> bytes:
    """Encode a long or int into a bytearray."""
    if not isinstance(value, int):
        raise EncoderException("Got non-int type for uvarint encoding: %s" % value)
    output = bytearray()
    if value < VarInt.MIN_UVARINT:
        raise EncoderException(
            "Error encoding %d as uvarint. Value must be positive" % value
        )
    if value > VarInt.MAX_UVARINT:
        raise EncoderException(
            "Error encoding %d as uvarint. Value must be %s or less"
            % (value, VarInt.MAX_UVARINT)
        )

    if not value:
        output.append(value & 0x7F)
    else:
        while value:
            next_byte = value & 0x7F
            value >>= 7
            if value:
                next_byte |= 0x80
            output.append(next_byte)

    return output

def decode_uvarint(buf: bytes, pos: int) -> Tuple[int, int]:
    """Decode bytearray into a long."""
    pos_start = pos
    # Convert buffer to string
    try:
        value = 0
        shift = 0
        while buf[pos] & 0x80:
            value += (buf[pos] & 0x7F) << (shift * 7)
            pos += 1
            shift += 1
        value += (buf[pos] & 0x7F) << (shift * 7)
        pos += 1
    except IndexError:
        raise DecoderException(
            "Error decoding uvarint: read past the end of the buffer"
        )

    # Validate that this is a cononical encoding by re-encoding the value
    try:
        test_encode = encode_uvarint(value)
    except EncoderException as ex:
        raise DecoderException(
            "Error decoding uvarint: value (%s) was not able to be re-encoded: %s"
            % (value, ex)
        )
    if buf[pos_start:pos] != test_encode:
        raise DecoderException(
            "Error decoding uvarint: Encoding is not standard:\noriginal:  %r\nstandard: %r"
            % (buf[pos_start:pos], test_encode)
        )

    return (value, pos)


def encode_varint(value: Any) -> bytes:
    """Encode a long or int into a bytearray."""
    if not isinstance(value, int):
        raise EncoderException("Got non-int type for varint encoding: %s" % value)
    if value > VarInt.MAX_SVARINT:
        raise EncoderException(
            "Error encoding %d as varint. Value must be <= %s" % (value, VarInt.MAX_SVARINT.value)
        )
    if value < VarInt.MIN_SVARINT:
        raise EncoderException(
            "Error encoding %d as varint. Value must be >= %s" % (value, VarInt.MIN_SVARINT.value)
        )
    if value < 0:
        value += 1 << 64
    output = encode_uvarint(value)
    return output


def decode_varint(buf: bytes, pos: int) -> Tuple[int, int]:
    """Decode bytearray into a long."""
    # Convert buffer to string
    pos_start = pos

    value, pos = decode_uvarint(buf, pos)
    if value & (1 << 63):
        value -= 1 << 64

    # Validate that this is a cononical encoding by re-encoding the value
    try:
        test_encode = encode_varint(value)
    except EncoderException as ex:
        raise DecoderException(
            "Error decoding varint: value (%s) was not able to be re-encoded: %s"
            % (value, ex)
        )

    if buf[pos_start:pos] != test_encode:
        raise DecoderException(
            "Error decoding varint: Encoding is not standard:\noriginal:  %r\nstandard: %r"
            % (buf[pos_start:pos], test_encode)
        )
    return (value, pos)


def encode_zig_zag(value: int) -> int:
    if value < 0:
        return (abs(value) << 1) - 1
    return value << 1


def decode_zig_zag(value: int) -> int:
    if value & 0x1:
        # negative
        return -((value + 1) >> 1)
    return value >> 1


def encode_svarint(value: Any) -> bytes:
    """Zigzag encode the potentially signed value prior to encoding"""
    if not isinstance(value, int):
        raise EncoderException("Got non-int type for svarint encoding: %s" % value)
    # zigzag encode value
    if value > VarInt.MAX_SVARINT:
        raise EncoderException(
            "Error encoding %d as svarint. Value must be <= %s" % (value, VarInt.MAX_SVARINT)
        )
    if value < VarInt.MIN_SVARINT:
        raise EncoderException(
            "Error encoding %d as svarint. Value must be >= %s" % (value, VarInt.MIN_SVARINT)
        )
    return encode_uvarint(encode_zig_zag(value))


def decode_svarint(buf: bytes, pos: int) -> Tuple[int, int]:
    """Decode bytearray into a long."""
    pos_start = pos

    output, pos = decode_uvarint(buf, pos)
    value = decode_zig_zag(output)

    # Validate that this is a cononical encoding by re-encoding the value
    test_encode = encode_svarint(value)
    if buf[pos_start:pos] != test_encode:
        raise DecoderException(
            "Error decoding svarint: Encoding is not standard:\noriginal:  %r\nstandard: %r"
            % (buf[pos_start:pos], test_encode)
        )

    return value, pos

def encode_string(value: Any) -> bytes:
    """Encode a string as a length delimited byte array"""
    try:
        if isinstance(value, bytes):
            value = value.decode("utf-8")
        elif not isinstance(value, str):
            raise TypeError("not expecting type '%s'" % type(value))
    except TypeError as exc:
        raise EncoderException("Error encoding string to message: %r" % value) from exc
    return encode_bytes(value)


def encode_bytes(value: Any) -> bytes:
    """Encode a length delimited byte array"""
    if isinstance(value, bytearray):
        value = bytes(value)
    try:
        if isinstance(value, str):
            value = value.encode("utf-8")
        if not isinstance(value, bytes):
            raise TypeError("not expecting type '%s'" % type(value))
    except TypeError as exc:
        raise EncoderException("Error encoding bytes to message: %r" % value) from exc

    if not isinstance(value, bytes):
        raise EncoderException(
            "encode_bytes must receive a bytes or bytearray value: %s %r"
            % (type(value), value)
        )
    encoded_length = encode_varint(len(value))
    return encoded_length + value


def decode_bytes(buf: bytes, pos: int) -> Tuple[bytes, int]:
    """Decode a length delimited bytes array from buf"""
    length, pos = decode_varint(buf, pos)
    end = pos + length
    try:
        return buf[pos:end], end
    except IndexError as exc:
        raise DecoderException(
            (
                "Error decoding bytes. Decoded length %d is longer than bytes"
                " available %d"
            )
            % (length, len(buf) - pos)
        ) from exc


def encode_bytes_hex(value: Any) -> bytes:
    """Encode a length delimited byte array represented by a hex string"""
    try:
        return encode_bytes(binascii.unhexlify(value))
    except (TypeError, binascii.Error) as exc:
        raise EncoderException("Error encoding hex bytestring %s" % value) from exc

def decode_bytes_hex(buf: bytes, pos: int) -> Tuple[bytes, int]:
    """Decode a length delimited byte array from buf and return a hex encoded string"""
    value, pos = decode_bytes(buf, pos)
    return binascii.hexlify(value), pos


def decode_string(value: bytes, pos: int) -> Tuple[str, int]:
    """Decode a length delimited byte array as a string"""
    length, pos = decode_varint(value, pos)
    end = pos + length
    try:
        # backslash escaping isn't reversible easily
        return value[pos:end].decode("utf-8"), end
    except (TypeError, UnicodeDecodeError) as exc:
        raise DecoderException("Error decoding UTF-8 string %r" % value[pos:end]) from exc


def encode_tag(field_number: int, wire_type: int) -> bytes:
    # Not checking bounds here, should be check before
    tag_number = (field_number << 3) | wire_type
    return encode_uvarint(tag_number)


def decode_tag(buf: bytes, pos: int) -> Tuple[int, int, int]:
    tag_number, pos = decode_uvarint(buf, pos)
    field_number = tag_number >> 3
    wire_type = tag_number & 0x7
    return field_number, wire_type, pos


from pydantic import BaseModel, ConfigDict, field_validator
import codecs
from pydantic_core.core_schema import FieldValidationInfo

def create_validated_model(extra_mode: str):
    class _ValidatedModel(BaseModel):
        model_config = ConfigDict(
            extra=extra_mode,
            validate_assignment=True,
            arbitrary_types_allowed=True
        )

        @field_validator("*", mode="before")
        @classmethod
        def check_special_fields(cls, v, info: FieldValidationInfo):
            if getattr(cls, "_encoding_fields", []) and info.field_name in cls._encoding_fields:
                if v is not None:
                    try:
                        codecs.lookup(v)
                    except LookupError:
                        raise ValueError(f"{info.field_name} got invalid encoding: {v}")
            if getattr(cls, "_path_fields", []) and info.field_name in cls._path_fields:
                if not isinstance(v, str) or not v.strip():
                    raise ValueError(f"{info.field_name} should be a non-empty string path")
            return v
    _ValidatedModel.__name__ = f"ValidatedModel_extra_{extra_mode}"
    return _ValidatedModel

BaseValidatedModel = create_validated_model("allow")
StrictValidatedModel = create_validated_model("ignore")

__all__ = [
    "BaseValidatedModel",
    "StrictValidatedModel",
]


import asyncio, inspect
from collections import defaultdict
# from ..utils import run_with_timeout
from typing import Set, TYPE_CHECKING, Callable, Any, TypeVar, Union, Awaitable
if TYPE_CHECKING:
    from ..crawler import Crawler
    from ..extensions import SignalInfo
    from ..settings import SettingsInfo
    from ..mq.kafka import KafkaManager

T = TypeVar("T")

class SignalManager:
    SignalCallback = Union[Callable[[T], Any], Callable[[T], Awaitable[Any]]]

    def __init__(self, stop_event=None, settings: "SettingsInfo"=None, maxsize=1000, kafkaManager: "KafkaManager"=None):
        self._listeners = defaultdict(list)
        self._queue = asyncio.Queue(maxsize=maxsize)
        self.stop_event: asyncio.Event = stop_event
        self._run_task = None
        self._pending_tasks: Set[asyncio.Task] = set()
        from ..utils import init_logger
        self.logger = init_logger(log_info=settings.LOG_INFO, logger_name=__name__)
        if kafkaManager:
            from ..utils import KafkaLoggingHandler
            kafka_handler = KafkaLoggingHandler(kafka=kafkaManager, stop_event=self.stop_event).create_fmt(settings)
            self.logger.addHandler(kafka_handler)

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event, 
            settings=crawler.settings,
            kafkaManager=crawler.kafkaManager,
        )

    def connect(self, signal: object, callback: SignalCallback):
        if not callable(callback):
            raise TypeError(f"Signal callback must be callable: got {type(callback)}")
        self._listeners[signal].append(callback)

    def send(self, signal: object, data: "SignalInfo"):
        if self.stop_event.is_set() or (not self._listeners[signal]):
            return
        asyncio.create_task(self._safe_put(signal, data))

    async def _safe_put(self, signal, data):
        if not self._listeners[signal]:
            return
        try:
            self._queue.put_nowait((signal, data))
        except asyncio.QueueFull:
            try:
                await asyncio.wait_for(self._queue.put((signal, data)), timeout=0.1)
            except asyncio.TimeoutError:
                self.logger.warning(f"[SignalManager] Signal queue full, dropped signal: {signal}")

    async def _dispatch(self, signal, data):
        for callback in self._listeners[signal]:
            try:
                if inspect.iscoroutinefunction(callback):
                    task = asyncio.create_task(callback(data))
                    self._pending_tasks.add(task)
                    task.add_done_callback(lambda t: self._pending_tasks.discard(t))
                else:
                    callback(data)
            except Exception as e:
                self.logger.error(f"[SignalManager] Signal callback error: {e}")

    async def run(self):
        try:
            while not self.stop_event.is_set():
                # signal, data = await run_with_timeout(self._queue.get, stop_event=self.stop_event, timeout=0.2)
                signal, data = await self._queue.get()
                if signal is not None:
                    await self._dispatch(signal, data)

                # try:
                #     signal, data = await asyncio.wait_for(self._queue.get(), timeout=3.0)
                #     if signal is not None:
                #         await self._dispatch(signal, data)
                # except asyncio.TimeoutError:
                #     pass
        except asyncio.CancelledError:
            self.logger.warning("[SignalManager] run task cancelled, waiting for pending callbacks...")
            for task in self._pending_tasks:
                task.cancel()
            await asyncio.sleep(0)

            try:
                await asyncio.wait_for(
                    asyncio.gather(*self._pending_tasks, return_exceptions=True),
                    timeout=3.0
                )
            except asyncio.TimeoutError:
                self.logger.warning("[SignalManager] Pending signal callbacks did not finish in time, force exit.")
            finally:
                raise

    def start(self):
        if not self._run_task or self._run_task.done():
            self._run_task = asyncio.create_task(self.run())

    async def stop(self):
        try:
            self._queue.put_nowait((None, None))
        except asyncio.QueueFull:
            pass

        if self._run_task:
            try:
                await asyncio.wait_for(self._run_task, timeout=3.0)
            except asyncio.TimeoutError:
                pass

        if self._pending_tasks:
            self.logger.info(f"[SignalManager] Waiting for {len(self._pending_tasks)} pending signal callback tasks to finish")
            await asyncio.gather(*self._pending_tasks, return_exceptions=True)


class BloomFilter:
    def __init__(self, size: int, expected: int, hash_count: int = 0): ...
    def add(self, data: str | bytes): ...
    def exists(self, data: str | bytes) -> bool: ...
    def free(self): ...
    def clear(self): ...
    def estimated_false_positive_rate(self) -> float: ...
    def get_hash_count(self) -> int: ...
    def get_indices(self, data: str | bytes) -> list[int]: ...


# typedef.py
from ..exceptions import TypedefException
from typing import TYPE_CHECKING, Dict, Optional, Any, List, Tuple, Union
if TYPE_CHECKING:
    from ..config import Config
    from ..pytypes import TypeDefDict, FieldDefDict


class TypeDef(object):
    def __init__(self):
        self._fields: Dict[str, "FieldDef"] = {}
        self._field_names: Dict[str, str] = {}

    @staticmethod
    def from_dict(typedef_dict: "TypeDefDict") -> "TypeDef":
        typedef = TypeDef()
        for field_id, fielddef_dict in typedef_dict.items():
            fielddef = FieldDef.from_dict(fielddef_dict, field_id)
            typedef._fields[field_id] = fielddef
            if fielddef._name:
                typedef._field_names[fielddef._name] = field_id
        return typedef

    def to_dict(self) -> "TypeDefDict":
        return {field: fielddef.to_dict() for field, fielddef in self._fields.items()}

    def make_mutable(self) -> "MutableTypeDef":
        mutable = MutableTypeDef()
        # Copy fields but don't deep copy
        # They should get immutable copies of FieldDef
        mutable._fields = self._fields.copy()
        return mutable

    def lookup_fielddef(self, key: str) -> Optional[Tuple[str, "FieldDef"]]:
        """Look up a field definition by number (w/ alt typedef) or name"""
        # We don't care what the alt number is now
        field_name = key.split("-", 1)[0]

        field_id = self._field_names.get(field_name, field_name)

        if field_id in self._fields:
            return field_id, self._fields[field_id]

        return None

    def lookup_fielddef_number(self, field_id: str) -> Optional[Tuple[str, "FieldDef"]]:
        if field_id in self._fields:
            return field_id, self._fields[field_id]
        return None

class MutableTypeDef(TypeDef):
    def set_fielddef(self, field_number: str, fielddef: "FieldDef") -> None:
        self._fields[field_number] = fielddef
        if fielddef._name:
            self._field_names[fielddef._name] = field_number

class FieldDef(object):
    def __init__(self, field_id: str) -> None:
        self._name: Optional[str] = None
        self._field_id: str = field_id
        self._message_type_name: Optional[str] = None
        # Normal type will be 0, alts starting at 1
        # Each field will have either the type or a typedef
        # We don't allow message_type_name in alt_typedefs, we'll use it for any "message" entry instead
        self._types: Dict[str, Union[str, "TypeDef"]] = {}
        self._example_value: Any = None
        self._seen_repeated: bool = False
        self._field_order: Optional[List[str]] = None

    @staticmethod
    def from_dict(fielddef_dict: "FieldDefDict", field_id: str) -> "FieldDef":
        fielddef = FieldDef(field_id)

        if "name" in fielddef_dict:
            fielddef._name = fielddef_dict["name"]

        if "type" in fielddef_dict:
            field_type = fielddef_dict["type"]

            if field_type == "message" and "message_typedef" in fielddef_dict:
                fielddef._types["0"] = TypeDef.from_dict(
                    fielddef_dict["message_typedef"]
                )
            else:
                fielddef._types["0"] = field_type

        if "message_type_name" in fielddef_dict:
            # We could lookup the typedef, but better to wait to resolve
            fielddef._message_type_name = fielddef_dict["message_type_name"]

        if "alt_typedefs" in fielddef_dict:
            for alt_number, alt_typedef in fielddef_dict["alt_typedefs"].items():
                if isinstance(alt_typedef, dict):
                    fielddef._types[alt_number] = TypeDef.from_dict(alt_typedef)
                else:
                    fielddef._types[alt_number] = alt_typedef

        if "example_value_ignored" in fielddef_dict:
            fielddef._example_value = fielddef_dict["example_value_ignored"]

        if "seen_repeated" in fielddef_dict:
            fielddef._seen_repeated = fielddef_dict["seen_repeated"]

        if "field_order" in fielddef_dict:
            fielddef._field_order = fielddef_dict["field_order"]

        return fielddef

    def to_dict(self) -> "FieldDefDict":
        fielddef_dict: "FieldDefDict" = {}
        if self._name:
            fielddef_dict["name"] = self._name
        if self._message_type_name:
            fielddef_dict["message_type_name"] = self._message_type_name
        if self._example_value:
            fielddef_dict["example_value_ignored"] = self._example_value
        if self._seen_repeated:
            fielddef_dict["seen_repeated"] = self._seen_repeated
        if self._field_order:
            fielddef_dict["field_order"] = self._field_order

        field_type = self._types.get("0")
        if isinstance(field_type, TypeDef):
            field_typedef = field_type.to_dict()
            field_type = "message"
            fielddef_dict["message_typedef"] = field_typedef

        if field_type:
            fielddef_dict["type"] = field_type

        if field_type and len(self._types) > 1:
            fielddef_dict["alt_typedefs"] = {
                alt_num: (
                    alt_type.to_dict() if isinstance(alt_type, TypeDef) else alt_type
                )
                for alt_num, alt_type in self._types.items()
                if alt_num != "0"
            }

        return fielddef_dict

    def make_mutable(self) -> "MutableFieldDef":
        mutable = MutableFieldDef(self._field_id)
        mutable._name = self._name
        mutable._message_type_name = self._message_type_name
        mutable._types = self._types.copy()
        mutable._example_value = self._example_value
        mutable._seen_repeated = self._seen_repeated
        mutable._field_order = self._field_order
        return mutable

    def lookup_field_type(self, key: str, config: "Config", field_path: List[str]) -> Optional[Union[str, "TypeDef"]]:
        if "-" in key:
            alt_type_id = key.split("-", 1)[1]
        else:
            alt_type_id = "0"

        return self.lookup_field_type_number(alt_type_id, config, field_path)

    # Lookup fieled type by just alt type number
    def lookup_field_type_number(self, alt_type_id: str, config: "Config", field_path: List[str]) -> Optional[Union[str, "TypeDef"]]:

        if alt_type_id not in self._types:
            return None

        field_type = self._types[alt_type_id]
        if field_type == "message":
            # We have to look up the message type name
            return self.resolve_message_type_name(config, field_path)

        return field_type

    @property
    def name(self) -> str:
        if self._name:
            return self._name
        return self._field_id

    def field_key(self, alt_field_id: str) -> str:
        if alt_field_id == "0":
            return self.name
        else:
            return self.name + "-" + alt_field_id

    def next_alt_type_id(self) -> str:
        existing_ids = [int(alt_type_id) for alt_type_id in self._types.keys()]
        if len(existing_ids) == 0:
            return "0"
        else:
            return str(max(existing_ids) + 1)

    @property
    def field_order(self) -> Optional[List[str]]:
        # someone could mutate the list returned, but that's not a high
        # priority
        return self._field_order

    @property
    def seen_repeated(self) -> bool:
        return self._seen_repeated

    def resolve_message_type_name(self, config: "Config", field_path: List[str]) -> "TypeDef":
        if self._message_type_name not in config.known_types:
            raise TypedefException(
                "Message name '%s' has not been defined in known types"
                % self._message_type_name,
                field_path,
            )
        return TypeDef.from_dict(config.known_types[self._message_type_name])

    def resolve_types(self, config: "Config", field_path: List[str]) -> Dict[str, Union[str, "TypeDef"]]:
        field_types = self._types.copy()
        if field_types.get("0") == "message":
            field_types["0"] = self.resolve_message_type_name(config, field_path)
        return field_types


class MutableFieldDef(FieldDef):
    def set_field_order(self, field_order: List[str]) -> None:
        self._field_order = field_order

    def mark_repeated(self) -> None:
        self._seen_repeated = True

    def set_type(self, alt_type_id: str, field_type: Union[str, "TypeDef"]) -> None:
        self._types[alt_type_id] = field_type

    def set_types(self, types: Dict[str, Union[str, "TypeDef"]]) -> None:
        self._types = types

    def add_type(self, field_type: Union[str, "TypeDef"]) -> str:
        alt_type_id = self.next_alt_type_id()
        self.set_type(alt_type_id, field_type)
        return alt_type_id



import asyncio
from scrapy_cffi.databases.mysql import SQLAlchemyMySQLManager
from sqlalchemy import select, Table, Column, Integer, String, MetaData
from sqlalchemy.ext.asyncio import AsyncSession

metadata = MetaData()

# Define a test table
user_table = Table(
    "test",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("name", String(50)),
)

async def test():
    # Initialize the MySQL manager
    await mysql.init()

    # 1. Create the table (run_sync is used to execute synchronous create_all)
    async with mysql.engine.begin() as conn:
        await conn.run_sync(metadata.create_all)

    # 2. Insert data
    async with mysql.session_factory() as session:
        session: AsyncSession
        await session.execute(user_table.insert().values(name="Alice"))
        await session.commit()

    # 3. Query data using run_stmt helper
    users = await mysql.run_stmt(select(user_table).where(user_table.c.name == "Alice"))
    print(users)

    # 3b. Query data directly with session
    async with mysql.session_factory() as session:
        stmt = select(user_table).where(user_table.c.name == "Alice")
        result = await session.execute(stmt)
        users = result.fetchall()
        print("查询结果：", users)

    # 4. Drop the table
    async with mysql.engine.begin() as conn:
        await conn.run_sync(metadata.drop_all)

    # 5. Close MySQL connection
    await mysql.close()

if __name__ == "__main__":
    mysql = SQLAlchemyMySQLManager(
        stop_event=asyncio.Event(),
        host="127.0.0.1",
        port=3306,
        db="test",
        user="root",
        password="123456"
    )
    asyncio.run(test())



import asyncio
from ..hooks import pipelines_hooks
from typing import TYPE_CHECKING, Union, Dict
if TYPE_CHECKING:
    from ..item import Item
    from ..crawler import Crawler
    from ..spiders import Spider
    from ..databases import RedisManager
    from ..databases.mysql import SQLAlchemyMySQLManager
    from ..databases.mongodb import MongoDBManager
    from ..settings import SettingsInfo
    from ..hooks.pipelines import PipelinesHooks
    from ..mq.rabbitmq import RabbitMQManager
    from ..mq.kafka import KafkaManager

class Pipeline:
    def __init__(
        self, 
        stop_event: asyncio.Event=None,
        settings: "SettingsInfo"=None, 
        redisManager: "RedisManager"=None, 
        mysqlManager: "SQLAlchemyMySQLManager"=None,
        mongodbManager: "MongoDBManager"=None,
        rabbitmqManager: "RabbitMQManager"=None,
        kafkaManager: "KafkaManager"=None,
        hooks: "PipelinesHooks"=None
    ):
        self.stop_event = stop_event
        self.settings = settings
        self.redisManager = redisManager
        self.mysqlManager = mysqlManager
        self.mongodbManager = mongodbManager
        self.rabbitmqManager = rabbitmqManager
        self.kafkaManager = kafkaManager
        self.hooks = hooks
        from ..utils import init_logger
        self.logger = init_logger(log_info=self.settings.LOG_INFO, logger_name=__name__)
        if kafkaManager:
            from ..utils import KafkaLoggingHandler
            kafka_handler = KafkaLoggingHandler(kafka=self.kafkaManager, stop_event=self.stop_event).create_fmt(self.settings)
            self.logger.addHandler(kafka_handler)

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event,
            settings=crawler.settings,
            redisManager=crawler.redisManager,
            mysqlManager=crawler.mysqlManager,
            mongodbManager=crawler.mongodbManager,
            rabbitmqManager=crawler.rabbitmqManager,
            kafkaManager=crawler.kafkaManager,
            hooks=pipelines_hooks(crawler)
        )

    async def open_spider(self, spider: "Spider"):
        pass

    async def process_item(self, item: Union["Item", Dict], spider: "Spider"):
        return item

    async def close_spider(self, spider: "Spider"):
        pass


# just a simple server for the demo project, without using the static dir.
try:
    import uvicorn, random, time, string
    from fastapi import FastAPI, Request, HTTPException, Response
    from fastapi.responses import PlainTextResponse, JSONResponse
except ImportError as e:
    raise ImportError(
        "Missing fastapi dependencies. "
        "Please install: pip install fastapi uvicorn"
    ) from e

app = FastAPI()

@app.exception_handler(HTTPException)
async def custom_http_exception_handler(request: Request, exc: HTTPException):
    return JSONResponse(
        status_code=exc.status_code,
        content={
            "data": [],
            "msg": exc.detail,
            "re_code": exc.status_code,
            "success": False,
        },
    )

def random_cookie():
    return ''.join(random.choices(string.ascii_letters + string.digits, k=50))

@app.get("/")
async def get_root(request: Request, response: Response):
    query_params = dict(request.query_params)
    response.set_cookie(key="demo_cookie", value=random_cookie(), httponly=True, path="/")
    return {"method": "GET", "params": query_params}

@app.get("/robots.txt", response_class=PlainTextResponse)
async def robot_url(request: Request):
    has_rebots = 1 if random.random() < 0.6 else 0
    if has_rebots:
        allow_all = 1 if random.random() < 0.6 else 0
        if allow_all:
            return ""
        return random.choice([
            "User-agent: *\nDisallow: /", 
            "User-agent: *\nDisallow: /teacher", 
            "User-agent: *\nDisallow: /student", 
            "User-agent: *\nDisallow: /files", 
        ])
    else:
        raise HTTPException(status_code=404, detail="not found")

def random_task_id():
    return int(str(time.time() + random.random() * 10 + random.random() * 100).replace(".", ""))

def random_str():
    return ''.join(random.choices(string.ascii_letters, k=random.randint(6, 20)))

@app.get("/task")
async def get_tasks(request: Request, response: Response):
    response.set_cookie(key="demo_cookie", value=random_cookie(), httponly=True, path="/")
    has_task = 1 if random.random() < 0.8 else 0
    task_type = request.query_params.get("type")
    if has_task:
        if task_type == "collect":
            small_task_type = random.choice(["school_collect", "teacher_collect", "student_collect"])
        elif task_type == "some_action":
            small_task_type = task_type
        else:
            return {"data": {"details": [], "platform": 0, "task_type": "task_type"}, "msg": "fail", "success": False}
        tasks_list = []
        tasks_count = random.randint(2, 5)
        # tasks_count = random.randint(10, 100)
        if small_task_type == "school_collect":
            task_key = "school_id"
        elif small_task_type == "teacher_collect":
            task_key = "teacher_id"
        elif small_task_type == "student_collect":
            task_key = "student_id"
        else:
            task_key = "action_id"
        for i in range(tasks_count):
            tasks_list.append({
                'small_task_id': random_task_id(), 
                'smail_task_type': small_task_type, 
                'small_task_info': {task_key: random.randint(30, 150), "extra_data": random_str()}
            })
        return {
            'data': {
                'details':tasks_list, 
                'platform': 0, 
                'task_id': random_task_id(), 
                'task_type': task_type
            }, 
            'msg': 'success', 're_code': 0, 'success': True
        }
    else:
        return {"data": {"details": [], "platform": 0, "task_type": task_type}, "msg": "success", "success": True}

@app.post("/update")
async def update_result(request: Request):
    try:
        data = await request.json()
        return {"data": {"details": [], "platform": 0}, "msg": "success", "success": True}
    except BaseException as e:
        return {"data": {"details": [], "platform": 0}, "msg": "fail", "success": False}

@app.get("/files/{img}", response_class=PlainTextResponse)
async def files_url(img: str, request: Request, response: Response):
    random_bytes = bytes([random.randint(0, 255) for _ in range(1024)])
    response.set_cookie(key="demo_cookie", value=random_cookie(), httponly=True, path="/")
    return Response(content=random_bytes, media_type="image/png")

def random_ids(start: int, end: int) -> list[int]:
    ids = []
    count = random.randint(start, end)
    for i in range(count):
        ids.append(i)
    return ids

@app.get("/school/{school_id}")
async def school_url(request: Request, school_id: int, response: Response):
    response.set_cookie(key="demo_cookie", value=random_cookie(), httponly=True, path="/")
    status = 1 if random.random() < 0.8 else 0
    if status and school_id > 0:
        has_next = 1 if random.random() < 0.6 else 0
        cursor = ""
        if has_next:
            cursor = f'{random_str()}{random_str()}{random_str()}'
        school_id += random.randint(100, 300)
        data_count = random.randint(10, int(str(school_id)[:2]))
        data_list = []
        for i in range(data_count):
            data_list.append({"class_id": i, "teacher_ids": random_ids(1, 5), "student_ids": random_ids(30, 50), "has_next": has_next, "cursor": cursor})
        return {"data": data_list, "msg": "success", "success": True}
    else:
        raise HTTPException(status_code=404, detail="not found")

@app.get("/teacher/{teacher_id}")
async def teacher_url(request: Request, teacher_id: int, response: Response):
    response.set_cookie(key="demo_cookie", value=random_cookie(), httponly=True, path="/")
    return {"teacher_id": teacher_id, "name": random_str(), "age": random.randint(6, 25), "gender": random.choice(["man", "lady"]), "img": f"/files/{random.randint(50000, 90000)}.png"}

@app.get("/student/{student_id}")
async def student_url(request: Request, student_id: int, response: Response):
    response.set_cookie(key="demo_cookie", value=random_cookie(), httponly=True, path="/")
    return {"student_id": student_id, "name": random_str(), "age": random.randint(6, 25), "gender": random.choice(["boy", "girl"]), "img": f"/files/{random.randint(50000, 90000)}.png"}

if __name__ == "__main__":
    uvicorn.run("fastApiServer:app", port=8002, reload=True) # debug=True


from ..core.downloader.internet import Request, WebSocketRequest, HttpResponse

class Failure(BaseException):
    def __init__(self, exception: BaseException):
        self.exception = exception

    def __str__(self):
        return f"<Failure exception={repr(self.exception)}>"

class RequestFailure(Failure):
    def __init__(self, exception: BaseException, request: Request):
        super().__init__(exception)
        self.request = request

    def __str__(self):
        return f"<{self.__class__.__name__} request={self.request.url} exception={repr(self.exception)}>"

class ResponseFailure(RequestFailure):
    def __init__(self, exception: BaseException, response: HttpResponse, request: Request):
        super().__init__(exception, request)
        self.response = response

    def __str__(self):
        return f"<{self.__class__.__name__} request={self.request.url} response.status_code={self.response.status_code}>"

class DownloadError(RequestFailure):
    pass

class IgnoreResponse(ResponseFailure):
    def __str__(self):
        return f"<IgnoreResponse request={self.request.url} non-200 response: response.status_code={self.response.status_code}>"

class ResponseError(ResponseFailure):
    pass

class SessionEndError(RequestFailure):
    def __str__(self):
        if isinstance(self.request, WebSocketRequest):
            return f"<SessionEndError websocket_id={self.request.websocket_id}: connection closed>"
        return f"<SessionEndError session_id={self.request.session_id}: session closed>"

class BlockRequestError(RequestFailure):
    def __str__(self):
        return f"<BlockRequestError request={self.request.url}: blocked by robots.txt>"

class FilterDomainRequestError(RequestFailure):
    def __str__(self):
        return f"<FilterDomainRequestError request={self.request.url}: not in allow_domains>"



import asyncio
import inspect
from tenacity import retry, wait_fixed, stop_after_attempt, retry_if_exception, RetryError
from functools import wraps
from typing import TYPE_CHECKING, Optional, Any, Callable
try:
    from sqlalchemy.ext.asyncio import create_async_engine, AsyncEngine, AsyncSession
    from sqlalchemy.exc import DBAPIError, OperationalError
    from sqlalchemy import text
    from sqlalchemy.orm import sessionmaker
except ImportError as e:
    raise ImportError(
        "Missing SQLAlchemy async dependencies. "
        "Please install: pip install sqlalchemy[asyncio] aiomysql"
    ) from e
if TYPE_CHECKING:
    from ..crawler import Crawler
    from sqlalchemy.sql import Executable

def is_fatal_error(msg: str) -> bool:
    return (
        "unknown database" in msg or
        "access denied" in msg or
        "doesn't exist" in msg
    )

def is_retryable_db_error(e: Exception) -> bool:
    if isinstance(e, (OperationalError, DBAPIError)):
        msg = str(e).lower()
        if is_fatal_error(msg):
            return False
        return True
    return False

def auto_retry(func: Callable):
    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        @retry(
            wait=wait_fixed(1),
            stop=stop_after_attempt(3),
            retry=retry_if_exception(is_retryable_db_error),
            reraise=True
        )
        async def _inner():
            if self.stop_event.is_set():
                raise asyncio.CancelledError("Stop event set, abort SQLAlchemy operation")
            try:
                return await func(self, *args, **kwargs)
            except (OperationalError, DBAPIError) as e:
                msg = str(e).lower()
                if is_fatal_error(msg):
                    print(f"[MySQL] error: {msg}")
                    self.stop_event.set()
                    raise asyncio.CancelledError("Fatal DB error, stopping all tasks")
                await self._reconnect()
                raise e
        try:
            return await _inner()
        except RetryError as e:
            print(f"[MySQL] reconnect failed：{e.last_attempt.exception()}")
            self.stop_event.set()
            raise e.last_attempt.exception()
    return wrapper

class SQLAlchemyMySQLManager:
    def __init__(self, stop_event, host, port, db, user, password):
        self.stop_event = stop_event
        self._db_url = f"mysql+aiomysql://{user}:{password}@{host}:{port}/{db}"
        self._method_cache = {}
        self.engine: Optional[AsyncEngine] = None
        self.session_factory = None

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event,
            host=crawler.settings.MYSQL_INFO.HOST,
            port=crawler.settings.MYSQL_INFO.PORT,
            user=crawler.settings.MYSQL_INFO.USERNAME,
            password=crawler.settings.MYSQL_INFO.PASSWORD,
            db=crawler.settings.MYSQL_INFO.DB,
        )

    async def init(self):
        await self._reconnect()

    async def _reconnect(self):
        if self.engine:
            await self.engine.dispose()

        self.engine = create_async_engine(self._db_url, echo=False, pool_pre_ping=True) # echo: debug log
        self.session_factory = sessionmaker(
            self.engine, class_=AsyncSession, expire_on_commit=False
        )

    def __getattribute__(self, name: str):
        if name.startswith("_") or name in ("_method_cache", "_reconnect", "stop_event"):
            return super().__getattribute__(name)
        attr = super().__getattribute__(name)
        if not callable(attr) or not inspect.iscoroutinefunction(attr):
            return attr
        method_cache = super().__getattribute__("_method_cache")
        if name not in method_cache:
            @wraps(attr)
            async def wrapper(*args, **kwargs):
                if self.stop_event.is_set():
                    raise asyncio.CancelledError(f"Stop event set, abort SQLAlchemy operation: {name}")
                try:
                    return await attr(*args, **kwargs)
                except (OperationalError, DBAPIError):
                    if self.stop_event.is_set():
                        raise asyncio.CancelledError("Stop event set during reconnect")
                    await self._reconnect()
                    return await attr(*args, **kwargs)
            method_cache[name] = wrapper
        return method_cache[name]

    @auto_retry
    async def execute(self, sql: str, params: Optional[dict[str, Any]] = None) -> None:
        """modify"""
        async with self.session_factory() as session:
            session: AsyncSession
            await session.execute(text(sql), params)
            await session.commit()

    @auto_retry
    async def fetchone(self, sql: str, params: Optional[dict[str, Any]] = None) -> Optional[Any]:
        """search one"""
        async with self.session_factory() as session:
            session: AsyncSession
            result = await session.execute(text(sql), params)
            return result.fetchone()

    @auto_retry
    async def fetchall(self, sql: str, params: Optional[dict[str, Any]] = None) -> list[Any]:
        """search all"""
        async with self.session_factory() as session:
            session: AsyncSession
            result = await session.execute(text(sql), params)
            return result.fetchall()
        
    @auto_retry
    async def run_stmt(self, stmt: "Executable", fetch: str = "all") -> Any:
        async with self.session_factory() as session:
            session: AsyncSession
            result = await session.execute(stmt)
            if fetch == "one":
                return result.fetchone()
            elif fetch == "scalar":
                return result.scalar()
            elif fetch == "scalars":
                return result.scalars().all()
            return result.fetchall()

    async def close(self):
        if self.engine:
            await self.engine.dispose()


import asyncio, threading, multiprocessing, inspect
from multiprocessing.managers import BaseManager as _BaseManager
from typing import Any, Callable, Union, Dict, List, Optional, Awaitable

# Start a new event loop in an async environment to run async code (this will occupy its own thread pool)
async def run_coroutine_in_new_loop(
    target: Union[Awaitable, Callable[..., Awaitable]],
    *args: Any,
    **kwargs: Any
) -> Any:
    def _runner():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            if inspect.isawaitable(target):
                coro = target
            elif callable(target):
                coro = target(*args, **kwargs)
                if not inspect.isawaitable(coro):
                    raise TypeError("Callable must return a coroutine")
            else:
                raise TypeError("target must be coroutine or coroutine-function")
            task = loop.create_task(coro)
            result = loop.run_until_complete(task)
            return result
        finally:
            pending = asyncio.all_tasks(loop)
            for t in pending:
                t.cancel()
            loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))
            loop.close()
    return await asyncio.to_thread(_runner)

# Start a new thread in an async environment to run async code
def run_coroutine_in_thread(
    target: Union[Awaitable, Callable[..., Awaitable]],
    *args: Any,
    **kwargs: Any
) -> asyncio.Future:
    loop = asyncio.get_running_loop()
    future = loop.create_future()

    def thread_worker():
        try:
            sub_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(sub_loop)
            if inspect.isawaitable(target):
                coro = target
            elif callable(target):
                coro = target(*args, **kwargs)
                if not inspect.isawaitable(coro):
                    raise TypeError("Callable must return a coroutine")
            else:
                raise TypeError("target must be coroutine or coroutine-function")
            
            result = sub_loop.run_until_complete(coro)
            loop.call_soon_threadsafe(future.set_result, result)
        except Exception as e:
            loop.call_soon_threadsafe(future.set_exception, e)
        finally:
            pending = asyncio.all_tasks(sub_loop)
            for task in pending:
                task.cancel()
            sub_loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))
            sub_loop.close()

    threading.Thread(target=thread_worker, daemon=True).start()
    return future

# Start a new process in an async environment to run async code (suitable for Linux/macOS)
# On Windows, due to process startup issues, Ctrl+C to interrupt will likely hang
# To be compatible with Windows startup, process_entrypoint cannot be inside a class or closure
def process_entrypoint(func: Callable, kwargs: Dict, queue: Optional[multiprocessing.Queue]):
    def handle_exit(sig, frame):
        print(f"[Child Process] Received signal {sig}, preparing to exit")
        import sys
        sys.exit(0)

    import platform
    if platform.system() != 'Windows':
        import os, signal
        try:
            os.setpgrp()
        except Exception as e:
            print(f"[Child Process] Failed to set process group: {e}")
        signal.signal(signal.SIGTERM, handle_exit)
        signal.signal(signal.SIGINT, handle_exit)

    try:
        if inspect.iscoroutinefunction(func):
            result = asyncio.run(func(**kwargs))
        else:
            result = func(**kwargs)
        if queue:
            queue.put((True, result))
    except Exception as e:
        print("Error: -----------------------------------------------------------------------------------------")
        if queue:
            queue.put((False, str(e)))
        else:
            print(f"[Detached Child Process Exception]: {e}")

class ProcessTaskManager:
    def __init__(self):
        import atexit
        self._procs: List[multiprocessing.Process] = []
        atexit.register(self.terminate_all)

    async def run(self, func: Callable, return_result=True, **kwargs):
        loop = asyncio.get_running_loop()
        if return_result:
            queue = multiprocessing.Queue()

            def start_proc():
                proc = multiprocessing.Process(
                    target=process_entrypoint,
                    args=(func, kwargs, queue)
                )
                proc.start()
                self._procs.append(proc)
                return proc, queue

            proc, queue = await loop.run_in_executor(None, start_proc)

            try:
                result = await loop.run_in_executor(None, queue.get)
                proc.join(timeout=3)
                if proc.is_alive():
                    proc.terminate()
                    proc.join()
                ok, val = result
                if ok:
                    return val
                raise RuntimeError(val)
            finally:
                if proc.is_alive():
                    proc.terminate()
                    proc.join()
        else:
            def start_detached():
                proc = multiprocessing.Process(
                    target=process_entrypoint,
                    args=(func, kwargs, None),
                    daemon=True
                )
                proc.start()
                self._procs.append(proc)

            await loop.run_in_executor(None, start_detached)

    def terminate_all(self):
        for proc in self._procs:
            if proc.is_alive():
                try:
                    proc.terminate()
                    proc.join(timeout=1)
                except Exception as e:
                    print(f"[Main Process] Failed to terminate child process: {e}")
        self._procs.clear()

class ProcessManager:
    """
    A general-purpose multiprocessing manager.
    - Supports registering functions, objects, and classes.
    - Provides server and client startup/shutdown.
    """
    class _Manager(_BaseManager): pass

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 50000,
        authkey: str = "abc",
        register_methods: Union[Dict[str, Any], None] = None,
    ):
        """
        :param host: Server host
        :param port: Server port
        :param authkey: Authentication key
        :param register_methods:
            - dict: { "name": callable / class / object }  -> for server registration
            - list or str: ["name1", "name2"] / "name"   -> for client registration only (placeholders)
        """
        self.address = (host, port)
        self.authkey = authkey.encode("utf-8")
        self._manager = None

        if isinstance(register_methods, dict):
            self._register_methods = register_methods
        elif isinstance(register_methods, list):
            self._register_methods = {name: None for name in register_methods}
        elif isinstance(register_methods, str):
            self._register_methods = {register_methods: None}
        else:
            self._register_methods = {}

    def _do_register(self):
        """Register functions, objects, or classes"""
        for name, target in self._register_methods.items():
            if target is None:
                self._Manager.register(name)
            elif isinstance(target, type):
                self._Manager.register(name, target)
            elif not callable(target):
                self._Manager.register(name, callable=lambda obj=target: obj)
            else:
                self._Manager.register(name, callable=target)

    def start_server(self, run_mode: int = 1):
        """Start a server process (background) or blocking mode"""
        self._do_register()
        self._manager = self._Manager(address=self.address, authkey=self.authkey)
        print(f"[ProcessManager] Server started at {self.address}")

        if run_mode:  # background mode
            self._manager.start()
        else:         # blocking mode
            server = self._manager.get_server()
            server.serve_forever()

    def shutdown(self):
        """Shutdown the server"""
        if self._manager:
            self._manager.shutdown()
            print("[ProcessManager] Server shutdown")

    def start_client(self):
        """Start a client and connect to the server"""
        import os
        self._do_register()
        self._manager = self._Manager(address=self.address, authkey=self.authkey)
        self._manager.connect()
        print(f"[ProcessManager] pid:{os.getpid()}, Client Connected to {self.address}")

    def __getattr__(self, item):
        if self._manager and hasattr(self._manager, item):
            return getattr(self._manager, item)
        raise AttributeError(item)
    
class ThreadFuture:
    def __init__(self, func: Callable, *args, **kwargs):
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self._result: Optional[Any] = None
        self._exception: Optional[Exception] = None
        self._done_event = threading.Event()
        self._callbacks: List[Callable[['ThreadFuture'], None]] = []

        self._thread = threading.Thread(target=self._run, daemon=True)
        self._thread.start()

    def _run(self):
        try:
            self._result = self._func(*self._args, **self._kwargs)
        except Exception as e:
            self._exception = e
        finally:
            self._done_event.set()
            for callback in self._callbacks:
                try:
                    callback(self)
                except Exception:
                    pass

    def done(self) -> bool:
        return self._done_event.is_set()

    def result(self, timeout: float = None) -> Any:
        finished = self._done_event.wait(timeout)
        if not finished:
            raise TimeoutError("Task not finished yet")
        if self._exception:
            raise self._exception
        return self._result

    def add_done_callback(self, fn: Callable[['ThreadFuture'], None]):
        if self.done():
            fn(self)
        else:
            self._callbacks.append(fn)

async def safe_call(
    func: Callable,
    *args,
    use_thread: bool = True,
    **kwargs
) -> Any:
    if inspect.iscoroutinefunction(func):
        return await func(*args, **kwargs)

    if use_thread:
        result = await asyncio.to_thread(func, *args, **kwargs)
    else:
        result = func(*args, **kwargs)

    if inspect.isawaitable(result):
        return await result
    return result


from pydantic import model_validator, Field
from enum import Enum
from typing import Optional, Union, List, Tuple
from .base import StrictValidatedModel

class BaseDBInfo(StrictValidatedModel):
    URL: Optional[str] = None
    HOST: Optional[str] = None
    PORT: Optional[Union[str, int]] = None
    USERNAME: Optional[str] = None
    PASSWORD: Optional[str] = None
    DB: Optional[Union[str, int]] = None

    @property
    def resolved_url(self) -> str:
        return self.URL if self.URL else None

class RedisInfo(BaseDBInfo):
    @model_validator(mode="after")
    def assemble_url(self) -> "RedisInfo":
        if not self.URL and self.HOST and self.PORT:
            auth_part = ""
            if self.USERNAME and self.PASSWORD:
                auth_part = f"{self.USERNAME}:{self.PASSWORD}@"
            elif self.PASSWORD:
                auth_part = f":{self.PASSWORD}@"
            db_part = f"/{self.DB}" if self.DB is not None else ""
            self.URL = f"redis://{auth_part}{self.HOST}:{self.PORT}{db_part}"
        return self

class RedisMode(str, Enum):
    SINGLE = "single"
    SENTINEL = "sentinel"
    CLUSTER = "cluster"
    
class RedisInfo(BaseDBInfo):
    MODE: Union[RedisMode, str] = RedisMode.SINGLE

    SENTINELS: Optional[List[tuple[str, int]]] = Field(default_factory=list)
    MASTER_NAME: Optional[str] = None  # sentinel mode
    SENTINEL_OVERRIDE_MASTER: Optional[Tuple[str, int]] = None # sentinel mode -> (master_host, master_port)

    CLUSTER_NODES: Optional[List[dict]] = Field(default_factory=list)

    @model_validator(mode="after")
    def assemble_url(self) -> "RedisInfo":
        if self.MODE == RedisMode.SINGLE:
            if not self.URL and self.HOST and self.PORT:
                auth_part = ""
                if self.USERNAME and self.PASSWORD:
                    auth_part = f"{self.USERNAME}:{self.PASSWORD}@"
                elif self.PASSWORD:
                    auth_part = f":{self.PASSWORD}@"
                db_part = f"/{self.DB}" if self.DB is not None else ""
                self.URL = f"redis://{auth_part}{self.HOST}:{self.PORT}{db_part}"
        return self

    @property
    def resolved_url(self) -> Union[str, List[tuple[str, int]], List[dict], None]:
        if self.MODE == RedisMode.SINGLE:
            return self.URL
        elif self.MODE == RedisMode.SENTINEL:
            return self.SENTINELS
        elif self.MODE == RedisMode.CLUSTER:
            return self.CLUSTER_NODES
        return None

class MysqlInfo(BaseDBInfo):
    DRIVER: str = "mysql+asyncmy" # default driver

    @model_validator(mode="after")
    def assemble_url(self) -> "MysqlInfo":
        if not self.URL and self.HOST and self.PORT:
            auth_part = ""
            if self.USERNAME and self.PASSWORD:
                auth_part = f"{self.USERNAME}:{self.PASSWORD}@"
            elif self.PASSWORD:
                auth_part = f":{self.PASSWORD}@"
            db_part = f"/{self.DB}" if self.DB is not None else ""
            self.URL = f"{self.DRIVER}://{auth_part}{self.HOST}:{self.PORT}{db_part}"
        return self

class MongodbInfo(BaseDBInfo):
    @model_validator(mode="after")
    def assemble_url(self) -> "MongodbInfo":
        if not self.URL and self.HOST and self.PORT:
            auth_part = ""
            if self.USERNAME and self.PASSWORD:
                auth_part = f"{self.USERNAME}:{self.PASSWORD}@"
            elif self.PASSWORD:
                auth_part = f":{self.PASSWORD}@"
            db_part = f"/{self.DB}" if self.DB is not None else ""
            self.URL = f"mongodb://{auth_part}{self.HOST}:{self.PORT}{db_part}"
        return self

__all__ = [
    "RedisInfo",
    "MysqlInfo",
    "MongodbInfo",
]


from scrapy_cffi.spiders.redis import RedisSpider
from scrapy_cffi.exceptions import Failure
from scrapy_cffi.internet import *

class CustomRedisSpider(RedisSpider):
    name = "customRedisSpider"
    robot_scheme = "http"
    allowed_domains = ["api.ipify.org", "127.0.0.1:8002", "localhost:8765"]
    redis_key = "customRedisSpider_test"
    count = 0
    'RPUSH customRedisSpider_test http://127.0.0.1:8002'

    async def parse(self, response: HttpResponse):
        print(response.session_id, response.text)
        yield WebSocketRequest(
            session_id=self.session_id,
            url="ws://localhost:8765",
            headers=self.settings.DEFAULT_HEADERS,
            cookies=self.settings.DEFAULT_COOKIES,
            proxies=self.settings.PROXIES,
            timeout=self.settings.TIMEOUT,
            dont_filter=self.settings.DONT_FILTER,
            callback=self.sec_test, 
            errback=self.errRet,
            send_message=WebSocketMsg(data=f"connect send test".encode('utf-8'))
        )

    async def sec_test(self, response: WebSocketResponse):
        print(f"spider received：{self.count}")
        if self.count < 3:
            print({"session_id": response.session_id, "data": response.msg[0].decode()})
            yield WebSocketRequest(
                session_id=self.session_id,
                websocket_id=response.websocket_id,
                send_message=WebSocketMsg(data=f"hello：{self.count}".encode('utf-8'))
            )
        elif self.count == 3:
            # scrapy_cffi version >= 0.2.0
            yield CloseSignal(session_id=self.session_id, websocket_end_for_key=response.websocket_id)

            # scrapy_cffi version 0.1.x
            # yield WebSocketRequest(
            #     session_id=self.session_id,
            #     websocket_id=response.websocket_id,
            #     websocket_end=True,
            #     # send_message=f"hello：{self.count}".encode('utf-8')
            # )
            yield {"session_id": response.session_id, "data": response.msg[0].decode()}
            yield {"session_id": response.session_id, "data": "spider end"}
        self.count += 1

    async def errRet(self, failure: Failure):
        print(str(failure))
        yield None


# models.py
from enum import Enum

class BaseWireTypes(int, Enum):
    VARINT = 0
    FIXED64 = 1
    LENGTH_DELIMITED = 2
    START_GROUP = 3
    END_GROUP = 4
    FIXED32 = 5

class VarInt(int, Enum):
    MAX_UVARINT = (1 << 64) - 1
    MIN_UVARINT = 0
    MAX_SVARINT = (1 << 63) - 1
    MIN_SVARINT = -(1 << 63)

class WireTypes(int, Enum):
    uint = BaseWireTypes.VARINT
    int = BaseWireTypes.VARINT
    sint = BaseWireTypes.VARINT
    fixed32 = BaseWireTypes.FIXED32
    sfixed32 = BaseWireTypes.FIXED32
    float = BaseWireTypes.FIXED32
    fixed64 = BaseWireTypes.FIXED64
    sfixed64 = BaseWireTypes.FIXED64
    double = BaseWireTypes.FIXED64
    bytes = BaseWireTypes.LENGTH_DELIMITED
    bytes_hex = BaseWireTypes.LENGTH_DELIMITED
    string = BaseWireTypes.LENGTH_DELIMITED
    message = BaseWireTypes.LENGTH_DELIMITED
    group = BaseWireTypes.START_GROUP
    packed_uint = BaseWireTypes.LENGTH_DELIMITED
    packed_int = BaseWireTypes.LENGTH_DELIMITED
    packed_sint = BaseWireTypes.LENGTH_DELIMITED
    packed_fixed32 = BaseWireTypes.LENGTH_DELIMITED
    packed_sfixed32 = BaseWireTypes.LENGTH_DELIMITED
    packed_float = BaseWireTypes.LENGTH_DELIMITED
    packed_fixed64 = BaseWireTypes.LENGTH_DELIMITED
    packed_sfixed64 = BaseWireTypes.LENGTH_DELIMITED
    packed_double = BaseWireTypes.LENGTH_DELIMITED

WIRE_TYPE_DEFAULTS = {
    BaseWireTypes.VARINT: "int",
    BaseWireTypes.FIXED32: "fixed32",
    BaseWireTypes.FIXED64: "fixed64",
    BaseWireTypes.LENGTH_DELIMITED: None,
    BaseWireTypes.START_GROUP: None,
    BaseWireTypes.END_GROUP: None,
} 



import base64
from typing import Optional, Dict, Any
from curl_cffi.const import CurlWsFlag
from pydantic import Field, field_validator
from .base import StrictValidatedModel
from ..utils import ProtobufFactory

class WebSocketMsg(StrictValidatedModel):
    data: Any = Field(default=b"")
    flags: Optional[CurlWsFlag] = Field(default=CurlWsFlag.BINARY)

    @field_validator("data", mode="before")
    @classmethod
    def ensure_bytes_and_infer_flag(cls, v, info):
        if isinstance(v, str):
            if not info.data.get("flags"):
                info.data["flags"] = CurlWsFlag.TEXT
            return v.encode("utf-8")
        elif isinstance(v, bytes):
            return v
        elif v is None:
            return b""
        raise TypeError(f"Unsupported data type for WebSocketMsg.data: {type(v)}")

    def to_dict(self) -> dict:
        return {
            "__wsmsg__": True,
            "data": base64.b64encode(self.data).decode(),
            "flags": int(self.flags or CurlWsFlag.BINARY),
        }

    @classmethod
    def from_dict(cls, d: dict) -> "WebSocketMsg":
        data = base64.b64decode(d["data"])
        flags = CurlWsFlag(d.get("flags", int(CurlWsFlag.BINARY)))
        return cls(data=data, flags=flags)

    def as_send_args(self) -> tuple[bytes, CurlWsFlag]:
        return self.data, self.flags or CurlWsFlag.BINARY

    def __repr__(self) -> str:
        f_name = getattr(self.flags, "name", str(self.flags))
        return f"<WebSocketMsg flags={f_name} len={len(self.data)}>"

    def protobuf_encode(self, typedef: Optional[Dict] = None):
        if typedef is None:
            return self
        self.data = ProtobufFactory.protobuf_encode(data=self.data, typedef=typedef)
        self.flags = CurlWsFlag.BINARY
        return self

    def grpc_encode(self, typedef: Optional[Dict] = None, is_gzip: bool = False):
        if typedef is None:
            return self
        self.data = ProtobufFactory.grpc_encode(data=self.data, typedef=typedef, is_gzip=is_gzip)
        self.flags = CurlWsFlag.BINARY
        return self



"""
RedisManager

An async Redis client extending `redis.asyncio.Redis` with full API support.

Features:
- Automatically retries and reconnects on connection failures.
- Respects a global asyncio stop event to gracefully abort operations during shutdown.
- Only allows certain Redis commands (e.g. DEL) to run when stopping to ensure safe cleanup.
- Provides convenience methods with built-in retry for common queue and deduplication patterns.

Designed for use within an asyncio event loop and single-threaded context.
"""
import json
import redis.asyncio as redis
from redis.exceptions import ConnectionError, TimeoutError
from tenacity import retry, wait_fixed, retry_if_exception_type
from functools import wraps
import inspect, asyncio
from typing import TYPE_CHECKING, Union, Tuple, List, Dict
if TYPE_CHECKING:
    from ..crawler import Crawler
    from redis.asyncio.client import Redis
    from redis.asyncio.connection import ConnectionPool

def auto_retry(func):
    @wraps(func)
    @retry(
        wait=wait_fixed(1),
        retry=retry_if_exception_type((ConnectionError, TimeoutError)),
        reraise=True
    )
    async def wrapper(self, *args, **kwargs):
        if self.stop_event.is_set():
            raise asyncio.CancelledError("Stop event set, abort Redis operation")
        try:
            return await func(self, *args, **kwargs)
        except (ConnectionError, TimeoutError):
            if self.stop_event.is_set():
                raise asyncio.CancelledError("Stop event set during reconnect")
            await self._reconnect()
            return await func(self, *args, **kwargs)
        except KeyboardInterrupt:
            pass
    return wrapper


class RedisManager(redis.Redis):
    def __init__(
        self,
        stop_event: asyncio.Event,
        redis_url: Union[str, List[Tuple[str, int]], List[Dict]],
        redis_mode: str = "single",
        master_name: str = None,
        sentinel_override_master: Tuple[str,int]=None,
        **kwargs
    ):
        self.stop_event = stop_event
        self.redis_mode = redis_mode
        self._redis_url = redis_url
        self._master_name = master_name
        self._sentinel_override_master = sentinel_override_master
        self._method_cache = {}
        self._sentinel = None

        if redis_mode == "single":
            tmp_instance: "Redis" = redis.from_url(redis_url, **kwargs)
        elif redis_mode == "sentinel":
            if not isinstance(redis_url, list):
                raise ValueError("Sentinel mode requires a list of (host, port)")
            from redis.sentinel import Sentinel
            self._sentinel = Sentinel(redis_url, **kwargs)
            if self._sentinel_override_master:
                host, port = self._sentinel_override_master
                tmp_instance = redis.Redis(host=host, port=port, **kwargs)
            else:
                tmp_instance = self._sentinel.master_for(master_name, **kwargs)
        elif redis_mode == "cluster":
            if not isinstance(redis_url, list):
                raise ValueError("Cluster mode requires a list of dict [{'host':..., 'port':...}] or list of URLs")
            from redis.cluster import RedisCluster

            if isinstance(redis_url[0], str):
                from urllib.parse import urlparse
                startup_nodes = [{"host": urlparse(u).hostname, "port": urlparse(u).port} for u in redis_url]
            else:
                startup_nodes = redis_url
            tmp_instance = RedisCluster(
                startup_nodes=startup_nodes,
                decode_responses=True,
                skip_full_coverage_check=True
            )
        else:
            raise ValueError(f"Unsupported redis_mode: {redis_mode}")

        super().__init__(
            connection_pool=tmp_instance.connection_pool,
            **{k: v for k, v in kwargs.items() if k in redis.Redis.__init__.__code__.co_varnames}
        )

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event,
            redis_mode=crawler.settings.REDIS_INFO.MODE,
            redis_url=crawler.settings.REDIS_INFO.resolved_url,
            master_name=crawler.settings.REDIS_INFO.MASTER_NAME,
            sentinel_override_master=crawler.settings.REDIS_INFO.SENTINEL_OVERRIDE_MASTER,
        )

    async def _reconnect(self):
        if self.stop_event.is_set():
            return
        await self.close()
        
        if self.redis_mode == "single":
            new_instance: "Redis" = redis.from_url(self._redis_url)
            self.connection_pool: "ConnectionPool" = new_instance.connection_pool
        elif self.redis_mode == "sentinel":
            if self._sentinel_override_master:
                host, port = self._sentinel_override_master
                new_instance = redis.Redis(host=host, port=port)
                self.connection_pool: "ConnectionPool" = new_instance.connection_pool
            else:
                master = self._sentinel.master_for(self._master_name)
                self.connection_pool: "ConnectionPool" = master.connection_pool
        elif self.redis_mode == "cluster":
            from redis.cluster import RedisCluster
            new_instance: RedisCluster  = RedisCluster(startup_nodes=self._redis_url)
            self.connection_pool: "ConnectionPool"  = new_instance.connection_pool

    def __getattribute__(self, name: str):
        if name.startswith("_") or name in ("_method_cache", "_reconnect"):
            return super().__getattribute__(name)

        attr = super().__getattribute__(name)

        if not callable(attr) or not inspect.iscoroutinefunction(attr):
            return attr

        method_cache = super().__getattribute__("_method_cache")

        if name not in method_cache:
            @wraps(attr)
            async def wrapper(*args, **kwargs):
                allowed_during_shutdown = {"execute_command", "initialize", "parse_response"}

                if self.stop_event.is_set():
                    if (name not in allowed_during_shutdown) or \
                        (name == "execute_command" and args[0] != "DEL") or \
                        (name == "parse_response" and args[1] != "DEL"):
                        raise asyncio.CancelledError(f"Stop event set, abort Redis operation: {name}")

                try:
                    if self.stop_event.is_set() and name in allowed_during_shutdown:
                        return await asyncio.wait_for(attr(*args, **kwargs), timeout=3)
                    else:
                        return await attr(*args, **kwargs)
                except (ConnectionError, TimeoutError):
                    if self.stop_event.is_set():
                        raise asyncio.CancelledError("Stop event set during reconnect")
                    await self._reconnect()
                    return await attr(*args, **kwargs)

            method_cache[name] = wrapper

        return method_cache[name]

    @auto_retry
    async def do_filter(self, fingerprint: str, key_new_seen: str, key_is_req: str):
        script = """
        local fingerprint = ARGV[1]
        if redis.call("SADD", KEYS[1], fingerprint) == 1 then
            if redis.call("SADD", KEYS[2], fingerprint) == 1 then
                redis.call("SREM", KEYS[2], fingerprint)
                return 1
            end
        end
        return 0
        """
        return await self.eval(
            script,
            2,
            key_new_seen,
            key_is_req,
            fingerprint
        )
    
    @auto_retry
    async def do_bloom_filter(
        self,
        key_new_seen: str,
        key_is_req: str,
        index_list: list[int]
    ) -> int:
        script = """
        local key_new_seen = KEYS[1]
        local key_is_req = KEYS[2]
        local indices = cjson.decode(ARGV[1])
        local is_new = 1

        for i=1,#indices do
            if redis.call("GETBIT", key_new_seen, indices[i]) == 0 then
                is_new = 1
                break
            else
                is_new = 0
            end
        end

        if is_new == 1 then
            for i=1,#indices do
                if redis.call("GETBIT", key_is_req, indices[i]) == 0 then
                    is_new = 1
                    break
                else
                    is_new = 0
                end
            end
        end

        if is_new == 1 then
            for i=1,#indices do
                redis.call("SETBIT", key_new_seen, indices[i], 1)
            end
        end

        return is_new
        """
        indices_json = json.dumps(index_list)
        return await self.eval(script, 2, key_new_seen, key_is_req, indices_json)

    @auto_retry
    async def dequeue_request(self, queue_key, timeout=2, decode_responses=False): # Pop a request from the queue, with optional timeout and decoding.
        result = await self.blpop(queue_key, timeout=timeout)
        if result:
            _, request = result
            if decode_responses:
                request: bytes
                request = request.decode('utf-8')
            return request
        return None


import tempfile, os
try:
    import magic
except ImportError as e:
    raise ImportError(
        "Missing magic dependencies. "
        "Please install one of the following according to your platform:\n"
        "  Windows: pip install scrapy_cffi[windows]\n"
        "  Linux/macOS: pip install scrapy_cffi[unix]"
    ) from e

try:
    from PIL import Image
except ImportError as e:
    raise ImportError(
        "Missing Pillow dependencies. "
        "Please install: pip install Pillow"
    ) from e

try:
    from hachoir.parser import createParser
    from hachoir.metadata import extractMetadata
except ImportError as e:
    raise ImportError(
        "Missing hachoir dependencies. "
        "Please install: pip install hachoir"
    ) from e
from typing import Union

def guess_content_type(byte_data: bytes) -> str:
    try:
        max_byte_index = 2048
        if len(byte_data) < max_byte_index:
            max_byte_index = len(byte_data)
        mime = magic.Magic(mime=True)
        mime_type = mime.from_buffer(byte_data[:max_byte_index])
        return mime_type
    except Exception as e:
        return f"Failed to guess: {e}"
    
def get_image_info_from_bytes(image_bytes: bytes) -> Union[dict, str]:
    try:
        from PIL import Image
        from io import BytesIO
        with Image.open(BytesIO(image_bytes)) as img:
            image_info = {
                "format": img.format,
                "mode": img.mode,
                "width": img.width,
                "height": img.height,
            }
            return image_info
    except Exception as e:
        return f"Failed to read image: {e}"
    
def get_video_info_from_bytes(video_bytes: bytes) -> Union[dict, str]:
    import subprocess, json
    try:
        subprocess.run(["ffprobe", "-version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
    except (FileNotFoundError, subprocess.CalledProcessError) as e:
        raise RuntimeError(
            "Missing ffprobe executable. Please install FFmpeg (which includes ffprobe) on your system:\n"
            "  Linux: sudo apt install ffmpeg\n"
            "  macOS: brew install ffmpeg\n"
            "  Windows: Download from https://ffmpeg.org/download.html and add to PATH"
        ) from e

    try:
        cmd = [
            "ffprobe",
            "-v", "error",
            "-print_format", "json",
            "-show_streams",
            "-show_format",
            "-"
        ]
        result = subprocess.run(cmd, input=video_bytes, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if result.returncode != 0:
            return f"ffprobe failed: {result.stderr.decode('utf-8')}"
        
        info = json.loads(result.stdout)
        video_stream = next((s for s in info.get("streams", []) if s.get("codec_type") == "video"), None)
        if not video_stream:
            return "No video stream found"

        width = int(video_stream.get("width", 0))
        height = int(video_stream.get("height", 0))
        duration = float(info.get("format", {}).get("duration", 0))

        return {"width": width, "height": height, "duration": duration}
    except Exception as e:
        return f"Failed to read video: {e}"
    
def get_image_info_from_tempfile(image_bytes: bytes) -> Union[dict, str]:
    with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as temp:
        temp.write(image_bytes)
        temp_path = temp.name

    try:
        with Image.open(temp_path) as img:
            image_info = {
                "format": img.format,
                "mode": img.mode,
                "width": img.width,
                "height": img.height,
            }
            return image_info
    except Exception as e:
        return f"Failed to read image: {e}"
    finally:
        try:
            os.remove(temp_path)
        except Exception as e:
            print(f"Failed to delete temporary file: {e}")

def get_video_info_from_tempfile(video_bytes: bytes) -> Union[dict, str]:
    with tempfile.NamedTemporaryFile(suffix=".mp4", delete=False) as temp:
        temp.write(video_bytes)
        temp_path = temp.name
    try:
        parser = createParser(temp_path)
        if not parser:
            return "Failed to parse video"
        metadata = extractMetadata(parser)
        if not metadata:
            return "Failed to extract video metadata"
        
        video_info = {}
        if metadata.has("duration"):
            video_info["duration"] = metadata.get("duration").total_seconds()
        else:
            return "Failed to get video duration"
        
        if metadata.has("width") and metadata.has("height"):
            video_info["width"] = metadata.get("width")
            video_info["height"] = metadata.get("height")
        else:
            return "Failed to get video resolution"
        return video_info
    finally:
        try:
            if parser and parser.stream:
                parser.stream.close()
        except Exception:
            pass
        try:
            os.remove(temp_path)
        except Exception as e:
            print(f"Failed to delete temporary file: {e}")



import json
from .base import MemoryDupeFilter
from ..databases import RedisManager
from ..core.downloader.internet import Request
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..spiders import Spider
    from ..settings import SettingsInfo
    from ..cpy.cpy_resources.bloom.fallback import BloomFilterPy

class RedisDupeFilter(MemoryDupeFilter):
    def __init__(self, settings: "SettingsInfo"=None, redisManager: RedisManager=None, **kwargs):
        super().__init__(settings=settings, **kwargs)
        self.new_seen = self.settings._NEW_SEEN # Requests marked as seen but not yet sent
        self.sent_seen = self.settings._SENT_SEEN # Requests that have been seen and already sent

        self.redisManager = redisManager
        if self.redisManager.redis_mode == "cluster":
            self.cluster_nodes = [f"{n['host']}:{n['port']}" for n in self.redisManager._redis_url]
        else:
            self.cluster_nodes = ["None"]

    async def request_seen(self, request: "Request", spider: "Spider") -> bool:
        # Requests with dont_filter=True or WebSocket requests signaling connection end should not be deduplicated
        if request.dont_filter:
            return False

        fingerprint = self.get_fingerprint(request=request)
        if self.redisManager.redis_mode == "cluster": # unsupported not SCHEDULER_PERSIST
            from ..utils import get_node
            node = get_node(self.cluster_nodes, fingerprint)
            key_new_seen = f"{self.new_seen}:{node}"
            key_is_req = f"{self.sent_seen}:{node}"
        else:
            key_new_seen = self.new_seen
            key_is_req = self.sent_seen

        is_new = await self.redisManager.do_filter(
            fingerprint=fingerprint,
            key_new_seen=key_new_seen,
            key_is_req=key_is_req,
        )
        if self.settings.DEDUP_TTL > 0:
            await self.redisManager.expire(key_new_seen, self.settings.DEDUP_TTL)
            await self.redisManager.expire(key_is_req, self.settings.DEDUP_TTL)
        return is_new == 0

    async def mark_sent(self, request: "Request", spider: "Spider", **kwargs):
        if not request.dont_filter:
            return await self.redisManager.sadd(self.sent_seen, self.get_fingerprint(request=request))
        
class RedisBloomDupeFilter(RedisDupeFilter):
    def __init__(self, settings: "SettingsInfo"=None, redisManager: RedisManager=None, **kwargs):
        super().__init__(settings=settings, redisManager=redisManager, **kwargs)
        import bloom
        self.bloomFilter: "BloomFilterPy" = bloom.BloomFilter(size=self.settings.BLOOM_INFO.SIZE, expected=self.settings.BLOOM_INFO.EXPECTED, hash_count=self.settings.BLOOM_INFO.HASH_COUNT)

    async def request_seen(self, request: "Request", spider: "Spider") -> bool:
        # Requests with dont_filter=True or WebSocket requests signaling connection end should not be deduplicated
        if request.dont_filter:
            return False

        origin_fp_bytes = self.create_bytes(request=request)
        index_list = self.bloomFilter.get_indices(origin_fp_bytes)
        if self.redisManager.redis_mode == "cluster": # unsupported not SCHEDULER_PERSIST
            from ..utils import get_node
            node = get_node(self.cluster_nodes, json.dumps(index_list, separators=(",", ":")))
            key_new_seen = f"{self.new_seen}:{node}"
            key_is_req = f"{self.sent_seen}:{node}"
        else:
            key_new_seen = self.new_seen
            key_is_req = self.sent_seen

        is_new = await self.redisManager.do_bloom_filter(
            key_new_seen=key_new_seen,
            key_is_req=key_is_req,
            index_list=index_list
        )
        if self.settings.DEDUP_TTL > 0:
            await self.redisManager.expire(key_new_seen, self.settings.DEDUP_TTL)
            await self.redisManager.expire(key_is_req, self.settings.DEDUP_TTL)
        return is_new == 0

    async def mark_sent(self, request: "Request", spider: "Spider", **kwargs):
        if not request.dont_filter:
            pipe = self.redisManager.pipeline()
            origin_fp_bytes = self.create_bytes(request=request)
            if self.redisManager.redis_mode == "cluster": # unsupported not SCHEDULER_PERSIST
                from ..utils import get_node
                node = get_node(self.cluster_nodes, origin_fp_bytes)
                key_is_req = f"{self.sent_seen}:{node}"
            else:
                key_is_req = self.sent_seen

            for idx in self.bloomFilter.get_indices(origin_fp_bytes):
                pipe.setbit(key_is_req, idx, 1)
            return await pipe.execute()


# type_maps.py
from ..utils import wire, fixed
from ..exceptions import DecoderException
from typing import Any, Callable, Dict, Tuple, List

def generate_packed_encoder(wrapped_encoder: Callable[[Any], bytes]) -> Callable[[List[Any]], bytes]:
    """Generate an encoder for a packed type based on a base type encoder"""

    def length_wrapper(values: List[Any]) -> bytes:
        # Encode repeat values and prefix with the length
        output = bytearray()
        for value in values:
            output += wrapped_encoder(value)
        length = wire.encode_varint(len(output))
        return length + output

    return length_wrapper

def generate_packed_decoder(wrapped_decoder: Callable[[bytes, int], Tuple[Any, int]]) -> Callable[[bytes, int], Tuple[List[Any], int]]:
    """Generate an decoder for a packed type based on a base type decoder"""

    def length_wrapper(buf: bytes, pos: int) -> Tuple[List[Any], int]:
        # Decode repeat values prefixed with the length
        length, pos = wire.decode_varint(buf, pos)
        end = pos + length
        output = []
        while pos < end:
            value, pos = wrapped_decoder(buf, pos)
            output.append(value)
        if pos > end:
            raise DecoderException(
                (
                    "Error decoding packed field. Packed length larger than"
                    " buffer: decoded = %d, left = %d"
                )
                % (length, len(buf) - pos)
            )
        return output, pos

    return length_wrapper

# Map a blackboxprotobuf type to specific encoder
ENCODERS: Dict[str, Callable[[Any], bytes]] = {
    "uint": wire.encode_uvarint,
    "int": wire.encode_varint,
    "sint": wire.encode_svarint,
    "fixed32": fixed.encode_fixed32,
    "sfixed32": fixed.encode_sfixed32,
    "float": fixed.encode_float,
    "fixed64": fixed.encode_fixed64,
    "sfixed64": fixed.encode_sfixed64,
    "double": fixed.encode_double,
    "bytes": wire.encode_bytes,
    "bytes_hex": wire.encode_bytes_hex,
    "string": wire.encode_string,
    "packed_uint": generate_packed_encoder(wire.encode_uvarint),
    "packed_int": generate_packed_encoder(wire.encode_varint),
    "packed_sint": generate_packed_encoder(wire.encode_svarint),
    "packed_fixed32": generate_packed_encoder(fixed.encode_fixed32),
    "packed_sfixed32": generate_packed_encoder(fixed.encode_sfixed32),
    "packed_float": generate_packed_encoder(fixed.encode_float),
    "packed_fixed64": generate_packed_encoder(fixed.encode_fixed64),
    "packed_sfixed64": generate_packed_encoder(fixed.encode_sfixed64),
    "packed_double": generate_packed_encoder(fixed.encode_double),
}

# Map a blackboxprotobuf type to specific decoder
DECODERS: Dict[str, Callable[[bytes, int], Tuple[Any, int]]] = {
    "uint": wire.decode_uvarint,
    "int": wire.decode_varint,
    "sint": wire.decode_svarint,
    "fixed32": fixed.decode_fixed32,
    "sfixed32": fixed.decode_sfixed32,
    "float": fixed.decode_float,
    "fixed64": fixed.decode_fixed64,
    "sfixed64": fixed.decode_sfixed64,
    "double": fixed.decode_double,
    "bytes": wire.decode_bytes,
    "bytes_hex": wire.decode_bytes_hex,
    "string": wire.decode_string,
    "packed_uint": generate_packed_decoder(wire.decode_uvarint),
    "packed_int": generate_packed_decoder(wire.decode_varint),
    "packed_sint": generate_packed_decoder(wire.decode_svarint),
    "packed_fixed32": generate_packed_decoder(fixed.decode_fixed32),
    "packed_sfixed32": generate_packed_decoder(fixed.decode_sfixed32),
    "packed_float": generate_packed_decoder(fixed.decode_float),
    "packed_fixed64": generate_packed_decoder(fixed.decode_fixed64),
    "packed_sfixed64": generate_packed_decoder(fixed.decode_sfixed64),
    "packed_double": generate_packed_decoder(fixed.decode_double),
}


from .sessions import SessionManager
from .engine import Engine
from .tasks import TaskManager
from .downloader import Downloader

__all__ = [
    "SessionManager",
    "Engine",
    "TaskManager",
    "Downloader"
]


import types
from ..models import StrictValidatedModel
from enum import Enum
from typing import Optional, TYPE_CHECKING, Union, Dict, List, Callable
from ..spiders import BaseSpider
from ..core.sessions import CloseSignal
from ..core.downloader.internet import Request
from ..core.downloader.internet import Response
from ..item import Item
if TYPE_CHECKING:
    from ..interceptors import SpiderInterceptor, DownloadInterceptor
    from ..databases import RedisManager
    from ..crawler import Crawler

class ChainNode:
    def __init__(self, instance):
        self.instance: Union["SpiderInterceptor", "DownloadInterceptor"] = instance
        self.prev: Union["SpiderInterceptor", "DownloadInterceptor"] = None
        self.next: Union["SpiderInterceptor", "DownloadInterceptor"] = None

class ChainManager:
    def __init__(self, crawler: "Crawler", class_list: list):
        self.chain_list: list[ChainNode] = []
        self.chain_head: Optional[ChainNode] = None
        self.chain_tail: Optional[ChainNode] = None
        self.create_chain(crawler, class_list)
        self.settings = crawler.settings
        self.redisManager: "RedisManager" = crawler.redisManager
        # from ..utils import init_logger
        # self.logger = init_logger(log_info=self.settings.LOG_INFO, logger_name=__name__)
        # if crawler.kafkaManager:
        #     from ..utils import KafkaLoggingHandler
        #     kafka_handler = KafkaLoggingHandler(kafka=crawler.kafkaManager, stop_event=crawler.stop_event).create_fmt(self.settings)
        #     self.logger.addHandler(kafka_handler)

    @classmethod
    def from_crawler(cls, crawler: "Crawler", class_list: list):
        return cls(
            crawler=crawler,
            class_list=class_list,
        )

    def create_chain(self, crawler: "Crawler", class_list: list["SpiderInterceptor", "DownloadInterceptor"]):
        nodes: List[ChainNode] = []
        for cls in class_list:
            inst = cls.from_crawler(crawler)
            nodes.append(ChainNode(inst))

        for i, node in enumerate(nodes):
            if i > 0: node.prev = nodes[i-1]
            if i < len(nodes) - 1: node.next = nodes[i+1]

        self.chain_head = nodes[0] if nodes else None
        self.chain_tail = nodes[-1] if nodes else None
        self.chain_list = nodes

    async def forward_pass(self, call_func_cls, call_func_name, pad_data=None, **kwargs):
        node: ChainNode = self.chain_head
        while node:
            call_func_addr = getattr(call_func_cls, call_func_name)
            result = await call_func_addr(pad_data, **kwargs)
            node = node.next
            if node:
                call_func_cls = node.instance
                if result is not None:
                    pad_data = result
        return result

    async def backward_pass(self, call_func_cls, call_func_name, item=None, **kwargs):
        node: ChainNode = self.chain_tail
        while node:
            call_func_addr = getattr(call_func_cls, call_func_name)
            result = await call_func_addr(item, **kwargs)
            node = node.prev
            if node:
                call_func_cls = node.instance
                if result is not None:
                    item = result
        return result
    
class ChainNextEnum(str, Enum):
    RESCHEDULE = "reschedule"
    RESPONSE = "response"
    EXCEPTION = "exception"
    DOWNLOADER = "downloader"
    PIPELINE = "pipeline"
    SPIDER = "spider"
    SESSION = "session"

class ChainResult(StrictValidatedModel):
    next: ChainNextEnum
    request: Optional[Request] = None
    response: Optional[Union[Response, BaseException]] = None
    spider: Optional[BaseSpider] = None
    item: Optional[Union[Item, Dict]] = None
    exception: Optional[BaseException] = None
    signal: Optional[CloseSignal] = None
    is_across: int = 0

async def _flatten_asyncgen(value):
    from collections.abc import AsyncIterable, Iterable
    ATOMIC_TYPES = (Request, Item, BaseException, CloseSignal, str, bytes, dict, type(None))
    if isinstance(value, AsyncIterable):
        async for v in value:
            async for sub_v in _flatten_asyncgen(v):
                yield sub_v
    elif (
        isinstance(value, Iterable)
        and not isinstance(value, ATOMIC_TYPES)
    ):
        for v in value:
            async for sub_v in _flatten_asyncgen(v):
                yield sub_v
    else:
        if not isinstance(value, ATOMIC_TYPES):
            raise TypeError(f"Unsupported spider return type: {type(value)}")
        yield value

def _ensure_asyncgen(result):
    if isinstance(result, types.CoroutineType):
        async def gen():
            value = await result
            async for item in _flatten_asyncgen(value):
                yield item
        return gen()
    else:
        async def gen():
            async for item in _flatten_asyncgen(result):
                yield item
        return gen()

# Interruptible chain
class InterruptibleChainManager(ChainManager):
    async def request_intercept_chain(self, request: Request, spider, callback: Callable):
        """
        Pass the request through the middleware chain from head to tail.

        For each middleware node:
        - If result is None, continue to next node.
        - If result is a Request, return ChainResult with next step RESCHEDULE and new request.
        - If result is a Response, return ChainResult with next step RESPONSE, passing current request and new response.
        - If result is an Exception, return ChainResult with next step EXCEPTION, passing exception and current request.
        - If result is invalid, raise ValueError.

        If all middleware return None, return ChainResult with next step DOWNLOADER to continue download phase.

        Returns:
            ChainResult directing the next step of processing.
        """
        node = self.chain_head
        while node:
            result = await node.instance.request_intercept(request=request, spider=spider)
            if result is None:
                node = node.next
                continue
            elif isinstance(result, Request):
                return await callback(ChainResult(next=ChainNextEnum.RESCHEDULE, request=result, spider=spider))
            elif isinstance(result, Response):
                return await callback(ChainResult(next=ChainNextEnum.RESPONSE, request=request, response=result, spider=spider))
            elif isinstance(result, BaseException):
                return await callback(ChainResult(next=ChainNextEnum.EXCEPTION, exception=result, request=request, spider=spider))
            raise callback(ValueError("request_intercept_chain got invalid return"))
        return await callback(ChainResult(next=ChainNextEnum.DOWNLOADER, request=request, spider=spider))
    
    async def response_intercept_chain(self, request: Request, response: Response, spider, callback: Callable):
        """
        Pass the response through the middleware chain from tail to head.

        For each middleware node:
        - If result is a Response, update response and continue to previous node.
        - If result is a Request, return ChainResult with next step RESCHEDULE and new request.
        - If result is an Exception, return ChainResult with next step EXCEPTION and exception.
        - If result is invalid, raise ValueError.

        If all middleware return None or Response, return ChainResult with next step SPIDER to proceed spider processing.

        Returns:
            ChainResult directing the next step of processing.
        """
        node = self.chain_tail
        while node:
            result = await node.instance.response_intercept(request=request, response=response, spider=spider)
            if result is not None:
                if isinstance(result, Response):
                    node = node.prev
                    response = result
                    continue
                elif isinstance(result, Request):
                    return await callback(ChainResult(next=ChainNextEnum.RESCHEDULE, request=result, spider=spider))
                elif isinstance(result, BaseException):
                    return await callback(ChainResult(next=ChainNextEnum.EXCEPTION, exception=result, request=request, spider=spider))
            raise ValueError("response_intercept_chain got invalid return")
        return await callback(ChainResult(next=ChainNextEnum.SPIDER, response=response, request=request, spider=spider))
    
    async def exception_intercept_chain(self, request: Request, exception: BaseException, spider, callback:  Callable, is_across=0):
        """
        Pass the exception through the middleware chain from tail to head for handling.

        For each middleware node:
        - If result is None, continue to previous node.
        - If result is an Exception, update exception and continue.
        - If result is a Request, return ChainResult with next step RESCHEDULE and new request.
        - If result is a Response, return ChainResult with next step SPIDER and new response.
        - If result is invalid, raise ValueError.

        If all middleware return None or Exception, return ChainResult with next step SPIDER with current exception.

        Args:
            is_across (int): flag indicating if the exception is propagated across middlewares.

        Returns:
            ChainResult directing the next step of processing.
        """
        node = self.chain_tail
        while node:
            result = await node.instance.exception_intercept(request=request, exception=exception, spider=spider)
            if result is None:
                node = node.prev
                continue
            elif isinstance(result, BaseException):
                node = node.prev
                exception = result
                continue
            elif isinstance(result, Request):
                return await callback(ChainResult(next=ChainNextEnum.RESCHEDULE, request=result, spider=spider, is_across=is_across))
            elif isinstance(result, Response):
                return await callback(ChainResult(next=ChainNextEnum.RESPONSE, response=result, request=request, spider=spider))
            raise ValueError("exception_intercept_chain got invalid return")
        return await callback(ChainResult(next=ChainNextEnum.SPIDER, response=result, request=request, spider=spider, is_across=is_across))
    
    async def process_spider_input_chain(self, response: Response, request: Request, spider, callback: Callable):
        """
        Pass the response through the spider input middleware chain sequentially.

        For each middleware node:
        - If the result is None, continue to the next node.
        - If the result is an exception, wrap it as a ChainResult with next step to spider's errback.
        - If the result is invalid (not None or Exception), raise an error.

        If all middlewares return None, wrap the original response as ChainResult directing to spider's callback.

        Returns:
            ChainResult object directing next processing step.
        """
        node = self.chain_head
        while node:
            result = await node.instance.process_spider_input(response=response, spider=spider)
            if result is None:
                node = node.next
                continue
            elif isinstance(result, BaseException):
                return await callback(ChainResult(next=ChainNextEnum.SPIDER, response=result, request=request, spider=spider)) # Pass to spider errback
            raise ValueError("process_spider_input_chain got invalid return")
        return await callback(ChainResult(next=ChainNextEnum.SPIDER, response=response, request=request, spider=spider)) # Pass to spider callback
    
    async def process_spider_output_chain(self, response: Response=None, result=None, spider=None):
        """
        Process the spider output through the middleware chain in reverse order (from tail to head).

        Workflow:
        1. Wrap the initial `result` into an async generator to uniformly handle single or multiple results.
        2. For each item yielded from the current results:
        - Pass it through each middleware node in reverse order.
        - At each node, call `process_spider_output` with the current item.
        - The middleware's output may be:
            - None: skip to the previous middleware node with the same item.
            - Single or multiple items: wrap as an async generator and iterate over them.
            - Each yielded item is processed similarly, ensuring a flat iteration.
        - If an Exception is yielded at any point, emit a ChainResult with EXCEPTION and stop processing.
        3. After passing through all middleware nodes, yield final ChainResult objects according to the processed item type:
        - Request: yield RESCHEDULE to retry the request.
        - Item or dict: yield PIPELINE to send the item downstream.
        - None: ignore.
        - Other unsupported types: raise ValueError.

        This design guarantees that the middleware chain can handle any combination of single or multiple results returned by each middleware,
        while keeping each middleware focused on processing individual items.
        """
        node = self.chain_tail
        while node:
            processed = await node.instance.process_spider_output(response=response, result=result, spider=spider)
            if processed is None:
                node = node.prev
                continue
            is_error = False
            async for i in _ensure_asyncgen(processed):
                if isinstance(i, (Request, Item, dict)):
                    result = i
                elif isinstance(i, BaseException):
                    result = i
                    yield ChainResult(next=ChainNextEnum.EXCEPTION, response=response, exception=result, spider=spider)
                    is_error = True
                    return
                elif isinstance(i, CloseSignal):
                    result = i
                    yield ChainResult(next=ChainNextEnum.SESSION, signal=result, spider=spider)
                    return
                else:
                    pass
                    # return f"Unsupported spider return type: {type(i)}"
            if not is_error:
                node = node.prev
        if result is None:
            pass
        elif isinstance(result, Request):
            yield ChainResult(next=ChainNextEnum.RESCHEDULE, request=result, spider=spider)
        elif isinstance(result, (Item, dict)):
            yield ChainResult(next=ChainNextEnum.PIPELINE, item=result, spider=spider)
        elif isinstance(result, CloseSignal):
            yield ChainResult(next=ChainNextEnum.SESSION, signal=result, spider=spider)
        else:
            raise ValueError(f"Unsupported spider return type: {type(result)}")
    
    async def process_spider_exception_chain(self, response: Response, exception: BaseException, spider, callback: Callable, is_across=0):
        """
        Pass the exception through the spider exception middleware chain from tail to head.

        For each middleware node:
        - If result is None, continue to previous node.
        - If result is an Exception, immediately return ChainResult with next step EXCEPTION, passing original response and exception.
        - If result is a Request, return ChainResult with next step RESCHEDULE and new request.
        - If result is an Item or dict, return ChainResult with next step PIPELINE and the item.
        - If result is invalid, raise ValueError.

        If all middleware return None, return ChainResult with next step EXCEPTION passing current response and exception.

        Args:
            is_across (int): flag indicating if the exception handling is propagated across middlewares.

        Returns:
            ChainResult directing the next step of processing.
        """
        node = self.chain_tail
        while node:
            result = await node.instance.process_spider_exception(response=response, exception=exception, spider=spider)
            if result is None:
                node = node.prev
                continue
            elif isinstance(result, BaseException):
                return await callback(ChainResult(next=ChainNextEnum.EXCEPTION, response=response, exception=exception, spider=spider, is_across=is_across))
            elif isinstance(result, Request):
                return await callback(ChainResult(next=ChainNextEnum.RESCHEDULE, request=result, spider=spider, is_across=is_across))
            elif isinstance(result, (Item, dict)):
                return await callback(ChainResult(next=ChainNextEnum.PIPELINE, item=result, spider=spider, is_across=is_across))
            raise ValueError("process_spider_exception_chain got invalid return")
        return await callback(ChainResult(next=ChainNextEnum.EXCEPTION, response=response, exception=exception, spider=spider, is_across=is_across))

__all__ = [
    "ChainManager",
    "InterruptibleChainManager"
]


# runner.py
import asyncio
import sys
import scrapy_cffi
from settings import create_settings
from typing import Tuple
if sys.platform.startswith("win"):
    # asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())
from scrapy_cffi.utils import setup_uvloop_once, get_or_create_loop
setup_uvloop_once()

# Ordinary users
def main(*args, **kwargs):
    settings = create_settings(spider_path="spiders.CustomSpider")

    # compatible scrapy settings.py
    # from scrapy_cffi import load_settings_with_path
    # settings = load_settings_with_path()

    scrapy_cffi.run_spider_sync(settings=settings, *args, **kwargs)

def main_all(*args, **kwargs):
    from scrapy_cffi.utils import get_run_py_dir
    spider_path = str(get_run_py_dir() / "spiders") # must be a directory when mode is 'run_all_spiders', since all spider files will be loaded from it
    settings = create_settings(spider_path=spider_path)

    # compatible scrapy settings.py
    # from scrapy_cffi import load_settings_with_path
    # settings = load_settings_with_path()
    
    scrapy_cffi.run_all_spiders_sync(settings=settings, *args, **kwargs)

# Advanced Users
async def advance_main(*args, **kwargs) -> Tuple[scrapy_cffi.crawler.Crawler, asyncio.Task]:
    settings = create_settings(spider_path="spiders.CustomSpider")

    # compatible scrapy settings.py
    # from scrapy_cffi import load_settings_with_path
    # settings = load_settings_with_path()

    crawler, engine_task = await scrapy_cffi.run_spider(settings=settings, new_loop=False, *args, **kwargs)
    return crawler, engine_task

async def advance_main_all(*args, **kwargs) -> Tuple[scrapy_cffi.crawler.Crawler, asyncio.Task]:
    from scrapy_cffi.utils import get_run_py_dir
    spider_path = str(get_run_py_dir() / "spiders") # must be a directory when mode is 'run_all_spiders', since all spider files will be loaded from it
    settings = create_settings(spider_path=spider_path)

    # compatible scrapy settings.py
    # from scrapy_cffi import load_settings_with_path
    # settings = load_settings_with_path()

    crawler, engine_task = await scrapy_cffi.run_all_spiders(settings=settings, new_loop=False, *args, **kwargs)
    return crawler, engine_task

# ————————————————————————————————————————————————————————————————————————
def setup_signal_handlers(loop: asyncio.AbstractEventLoop, shutdown_event: asyncio.Event):
    if sys.platform == "win32":
        print(">>> [info] Signal handlers not supported on Windows (fallback to KeyboardInterrupt)")
        return

    import signal
    def _handle_signal():
        print(">>> [signal] Received stop signal")
        shutdown_event.set()

    for sig in (signal.SIGINT, signal.SIGTERM):
        try:
            loop.add_signal_handler(sig, _handle_signal)
        except (NotImplementedError, ValueError):
            pass

if __name__ == "__main__":
    loop = get_or_create_loop()
    shutdown_event = asyncio.Event()
    setup_signal_handlers(loop, shutdown_event)

    crawler: scrapy_cffi.crawler.Crawler = None

    async def demo_main():
        global crawler
        # crawler, engine_task = await advance_main()
        crawler, engine_task = await advance_main_all()

        done, _ = await asyncio.wait(
            [engine_task, asyncio.create_task(shutdown_event.wait())],
            return_when=asyncio.FIRST_COMPLETED
        )

        if shutdown_event.is_set():
            print(">>> [main] Triggered shutdown, cleaning up...")
        else:
            print(">>> [main] Task finished normally.")

        await crawler.shutdown()

    try:
        loop.run_until_complete(demo_main())
    except KeyboardInterrupt:
        print(">>> [KeyboardInterrupt] manual stop")
        if crawler:
            loop.run_until_complete(crawler.shutdown())
    finally:
        scrapy_cffi.cleanup_loop(loop=loop)


    # ————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————
    # Ordinary user (internal automatic new loop)
    import threading
    t = threading.Thread(target=main)
    t.start()
    t.join()
    t = threading.Thread(target=main_all)
    t.start()
    t.join()


import asyncio
from .base import Pipeline
from typing import TYPE_CHECKING, Union, Dict
from ..hooks import _pipelines_hooks
if TYPE_CHECKING:
    from ..item import Item
    from ..crawler import Crawler
    from ..spiders import Spider
    from ..hooks.pipelines import _PipelinesHooks
    from ..databases import RedisManager
    from ..databases.mysql import SQLAlchemyMySQLManager
    from ..databases.mongodb import MongoDBManager
    from ..mq.rabbitmq import RabbitMQManager
    from ..mq.kafka import KafkaManager
    from ..settings import SettingsInfo

class _InnerPipeline(Pipeline): # scrapy_cffi version 0.1.x
    def __init__(
        self, 
        stop_event: asyncio.Event=None,
        settings: "SettingsInfo"=None, 
        redisManager: "RedisManager"=None, 
        mysqlManager: "SQLAlchemyMySQLManager"=None,
        mongodbManager: "MongoDBManager"=None,
        rabbitmqManager: "RabbitMQManager"=None,
        kafkaManager: "KafkaManager"=None,
        hooks: "_PipelinesHooks"=None
    ):
        super().__init__(
            stop_event=stop_event,
            settings=settings, 
            redisManager=redisManager,
            mysqlManager=mysqlManager,
            mongodbManager=mongodbManager,
            rabbitmqManager=rabbitmqManager,
            kafkaManager=kafkaManager,
        )
        self.hooks = hooks

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event,
            settings=crawler.settings,
            redisManager=crawler.redisManager,
            mysqlManager=crawler.mysqlManager,
            mongodbManager=crawler.mongodbManager,
            rabbitmqManager=crawler.rabbitmqManager,
            kafkaManager=crawler.kafkaManager,
            hooks=_pipelines_hooks(crawler)
        )

    async def open_spider(self, spider: "Spider"):
        pass

    async def process_item(self, item: Union["Item", Dict], spider: "Spider"):
        if self.settings.RET_COOKIES:
            if item.get("session_end"):
                self.hooks.session.mark_end(item.get("session_id"))
            ret_cookies = self.hooks.session.get_session_cookies(session_id=item.get("session_id"))
            item[self.settings.RET_COOKIES] = ret_cookies
        self.logger.debug(f"[PIPELINE] Processing item: {item}")
        return item

    async def close_spider(self, spider: "Spider"):
        # self.logger.debug(f"[PIPELINE] {spider.__class__.name} closed.")
        pass



import toml
from pathlib import Path
from jinja2 import Template

def check_config(project_path: Path, use_redis: bool=False, use_rabbitmq: bool=False, use_kafka: bool=False):
    config_path = project_path / "scrapy_cffi.toml"

    config_data = toml.load(config_path)
    if config_data.get("default"):
        if use_rabbitmq and (not config_data["default"].get("use_rabbitmq", False)):
            config_data["default"]["use_rabbitmq"] = True
            with config_path.open("w", encoding="utf-8") as f:
                toml.dump(config_data, f)

            # Scheduler -> RabbitMqScheduler
            settings_file = project_path / "settings.py"
            settings_data = settings_file.read_text(encoding='utf-8')
            if 'used_rabbitmq=False,' in settings_data:
                settings_data = settings_data.replace('used_rabbitmq=False,', 'used_rabbitmq=True,')
                settings_file.write_text(settings_data, encoding="utf-8")
        elif use_redis and (not config_data["default"].get("use_redis", False)):
            config_data["default"]["use_redis"] = True
            with config_path.open("w", encoding="utf-8") as f:
                toml.dump(config_data, f)

            # Scheduler -> RedisScheduler
            settings_file = project_path / "settings.py"
            settings_data = settings_file.read_text(encoding='utf-8')
            if 'used_redis=False,' in settings_data:
                settings_data = settings_data.replace('used_redis=False,', 'used_redis=True,')
                settings_file.write_text(settings_data, encoding="utf-8")

        if use_kafka and (not config_data["default"].get("use_kafka", False)):
            config_data["default"]["use_kafka"] = True
            with config_path.open("w", encoding="utf-8") as f:
                toml.dump(config_data, f)

            settings_file = project_path / "settings.py"
            settings_data = settings_file.read_text(encoding='utf-8')
            if 'used_kafka=False,' in settings_data:
                settings_data = settings_data.replace('used_kafka=False,', 'used_kafka=True,')
                settings_file.write_text(settings_data, encoding="utf-8")

def run(spider_name: str, allow_domain: str, use_redis: bool, use_rabbitmq: bool, use_kafka: bool, is_demo=False):
    from .base import find_project_root
    project_path = find_project_root()
    check_config(project_path, use_redis, use_rabbitmq, use_kafka)

    class_name = snake_to_camel(spider_name)

    if use_rabbitmq:
        import_path = "from scrapy_cffi.spiders.rabbitmq import RabbitmqSpider"
        base_class = "RabbitmqSpider"
        start_urls = 'rabbitmq_queue = ""'
    elif use_redis:
        import_path = "from scrapy_cffi.spiders.redis import RedisSpider"
        base_class = "RedisSpider"
        start_urls = 'redis_key = ""'
    else:
        import_path = "from scrapy_cffi.spiders import Spider"
        base_class = "Spider"
        start_urls = f'start_urls = ["https://{allow_domain}"]'

    base = Path(__file__).parent.parent # scrapy_cffi
    template_dir = base / "templates"
    with open(template_dir / "spider.py.j2", "r", encoding="utf-8") as f:
        template: Template = Template(f.read())
    
    code = template.render(
        class_name=class_name,
        spider_name=spider_name,
        domain=allow_domain,
        import_path=import_path,
        base_class=base_class,
        start_urls=start_urls
    )
    target_file = project_path / "spiders" / f"{spider_name}.py" # use abspath
    target_file.parent.mkdir(parents=True, exist_ok=True)
    target_file.write_text(code, encoding="utf-8")

    # To avoid overwriting user-defined content, only spider templates should be regenerated; other files should be appended or updated dynamically.
    update_spiders_init(project_path, class_name, spider_name)
    if not is_demo:
        print(f"Spider created: {target_file}")

# Use this to automatically convert snake_case to camelCase.
def snake_to_camel(name: str) -> str:
    return ''.join(word.capitalize() for word in name.split('_')) + "Spider"

# auto import
def update_spiders_init(project_path: Path, class_name: str, spider_name: str):
    init_path = project_path / "spiders" / "__init__.py"
    import_line = f"from .{spider_name} import {class_name}\n"

    if not init_path.exists():
        init_path.write_text(import_line, encoding="utf-8")
        return

    init_data = init_path.read_text(encoding='utf-8')
    if import_line in init_data:
        return
    with open(init_path, "a", encoding="utf-8") as f:
        f.write(import_line)